#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Module that contins various DNS constant definitions
"""


from datetime import datetime
import re
from socket import socket
from socket import AF_INET
from socket import AF_INET6
from socket import error as socket_error
from socket import inet_pton
from socket import inet_ntop

import dns.name
import dns.ttl

from magcode.core.database import db_time
from dms.exceptions import HostnameZiParseError as _HostnameZiParseError
from dms.exceptions import TtlZiParseError as _TtlZiParseError
from dms.exceptions import SOASerialArithmeticError \
                                    as _SOASerialArithmeticError
from dms.exceptions import SOASerialPublishedError \
                                    as _SOASerialPublishedError
from dms.exceptions import SOASerialOcclusionError \
                                    as _SOASerialOcclusionError


# Resource Record Classes
RRCLASS_IN = 'IN'
RRCLASS_HS = 'HS'
RRCLASS_CH = 'CH'

# Resource Record types
RRTYPE_NULL = 'NULL'
RRTYPE_A = 'A'
RRTYPE_AAAA = 'AAAA'
RRTYPE_CERT = 'CERT'
RRTYPE_CNAME = 'CNAME'
RRTYPE_DNAME = 'DNAME'
RRTYPE_DNSKEY = 'DNSKEY'
RRTYPE_DS = 'DS'
RRTYPE_HINFO = 'HINFO'
RRTYPE_IPSECKEY = 'IPSECKEY'
RRTYPE_KEY = 'KEY'
RRTYPE_KX = 'KX'
RRTYPE_LOC = 'LOC'
RRTYPE_MX = 'MX'
RRTYPE_NAPTR = 'NAPTR'
RRTYPE_NSAP = 'NSAP'
RRTYPE_NSEC = 'NSEC'
RRTYPE_NSEC3 = 'NSEC3'
RRTYPE_NSEC3PARAM = 'NSEC3PARAM'
RRTYPE_NS = 'NS'
RRTYPE_NXT = 'NXT'
RRTYPE_PTR = 'PTR'
RRTYPE_RP = 'RP'
RRTYPE_RRSIG = 'RRSIG'
RRTYPE_SIG = 'SIG'
RRTYPE_SOA = 'SOA'
RRTYPE_SPF = 'SPF'
RRTYPE_SRV = 'SRV'
RRTYPE_SSHFP = 'SSHFP'
RRTYPE_TXT = 'TXT'
RRTYPE_ANY = 'ANY'
RRTYPE_TLSA = 'TLSA'

# Various RDATA representations
RDATA_GENERIC_NULL = '\# 0'

# Update operations
RROP_ADD = 'ADD'
RROP_UPDATE_RRTYPE = 'UPDATE_RRTYPE'
RROP_DELETE = 'DELETE'
RROP_PTR_UPDATE = 'PTR_UPDATE'
RROP_PTR_UPDATE_FORCE = 'PTR_UPDATE_FORCE'
# This does not include the PTR_UPDATE update operations for security reasons
# They can be tested anyhow by using A and AAAA records in a forward zone
rr_op_values = (None, RROP_ADD, RROP_UPDATE_RRTYPE, RROP_DELETE)

# Domain name element constants
DOMN_MAXLEN = 255
DOMN_LBLLEN = 63
DOMN_CHRREGEXP = r'^[-.a-zA-Z0-9]*$'
DOMN_LBLREGEXP = r'^$|^[a-zA-Z0-9][-a-zA-Z0-9]*$'
DOMN_IPV6_REGEXP = r'^[a-fA-F0-9:]+$'
DOMN_IPV4_REGEXP = r'^[0-9.]+$'
DOMN_LBLSEP = '.'

label_re = re.compile(DOMN_LBLREGEXP)
ipv6_re = re.compile(DOMN_IPV6_REGEXP)
ipv4_re = re.compile(DOMN_IPV4_REGEXP)

def is_inet_domain(text):
    if len(text) > DOMN_MAXLEN:
        return False
    labels = text.split(DOMN_LBLSEP)
    if labels[-1] != '':
        #Must be root domain
        return False
    for lbl in labels[:-1]:
        if not lbl:
            # label must be at least one character long
            return False
        if len(lbl) > DOMN_LBLLEN:
            return False
        if not label_re.search(lbl):
            return False
        if lbl[0] == '-' or lbl[-1] == '-':
            return False
    return True

def is_inet_hostname(text, absolute=False, wildcard=True):
    # Deal with special case '@'
    if text == '@':
        return True
    if len(text) > DOMN_MAXLEN:
        return False
    labels = text.split(DOMN_LBLSEP)
    if absolute and labels[-1] != '':
        # root domain must be appended
        return False
    lbl_list = labels[:-1] if labels[-1] == '' else labels
    for lbl in lbl_list:
        if not lbl:
            # label must be at least one character long
            return False
        # Wild card domain
        if (wildcard and lbl == '*'):
            continue
        if len(lbl) > DOMN_LBLLEN:
            return False
        if not label_re.search(lbl):
            return False
        if lbl[0] == '-' or lbl[-1] == '-':
            return False
    return True

def validate_zi_hostname(name, zi_field, text):
    """
    Validate that a hostname is 100%
    """
    try:
        thing = dns.name.from_text(text)
    except Exception as exc:
        raise _HostnameZiParseError(name, zi_field, text, str(exc))
    if not is_inet_hostname(text):
        raise _HostnameZiParseError(name, zi_field, text, None)

def validate_zi_ttl(name, zi_field, text):
    """
    Validate a ttl value
    """
    if len(text) > 20:
        raise _TtlZiParseError(name, zi_field, text, "longer than 20 chars.")
    try:
        thing = dns.ttl.from_text(text)
    except Exception as exc:
        raise _TtlZiParseError(name, zi_field, text, str(exc))

def new_zone_soa_serial(db_session):
    """
    Generate a new SOA serial number based on date, for initialising zones
    """
    date = db_time(db_session).timetuple()
    soa_serial = (00 + 100 * date.tm_mday + 10000 * date.tm_mon
                        + 1000000 * date.tm_year)
    return soa_serial

def is_network_address(address):
    """
    Vaildate a network address
    """
    # Check routines also over in dms.zone_tool if this needs to be
    # changed
    try:
        inet_pton(AF_INET6, address)
        return True
    except socket_error:
        pass
    try:
        inet_pton(AF_INET, address)
        return True
    except socket_error:
        pass
    return False

def split_cidr_network_tuple(cidr_network, filter_mask_size=True):
    """
    Split a CIDR network/mask to a values suitable for a DNS reverse zone
    """
    # Check routines also over in dms.zone_tool if this needs to be
    # changed
    try:
        network, mask = cidr_network.split('/')
        mask = int(mask)
    except ValueError:
        return ()
    if network.find(':') >= 0 and network.find('.') < 0:
        try:
            i_net = int.from_bytes(inet_pton(AF_INET6, network), 
                                byteorder='big',
                                signed=False)
            if filter_mask_size and mask not in range(4, 65, 4):
                return ()
            i_mask = ~(2**(128-mask)-1)
            i_net = i_net & i_mask
            network = inet_ntop(AF_INET6, 
                        i_net.to_bytes(16, byteorder='big', signed=False))
            return (network, mask)
        except socket_error:
            pass
    elif network.isdigit() or network.find('.') >= 0 and network.find(':') < 0:
        try:
            network = network[:-1] if network.endswith('.') else network
            num_bytes = len(network.split('.'))
            if num_bytes < 4:
                network += (4 - num_bytes) * '.0'
            if filter_mask_size and mask not in range(8, 25, 8):
                return ()
            i_net = int.from_bytes(inet_pton(AF_INET, network),
                                byteorder='big',
                                signed=False)
            i_mask = ~(2**(32-mask)-1)
            i_net = i_net & i_mask
            network = inet_ntop(AF_INET,
                        i_net.to_bytes(4, byteorder='big', signed=False))
            return (network, mask)
        except socket_error:
            pass
    return ()

def wellformed_cidr_network(cidr_network, filter_mask_size=True):
    """
    Produced a well-formed network/mask pair
    """
    result = split_cidr_network_tuple(cidr_network, filter_mask_size)
    return '%s/%s' % result if result else ''

def zone_name_from_network(cidr_network):
    """
    Convert a CIDR network address to a reverse zone name

    Partly inspired by dnspython dns.reverse.from_address()
    """
    # Check routines also over in dms.zone_tool if this needs to be
    # changed
    result = split_cidr_network_tuple(cidr_network)
    if not result:
        return ()
    network, mask = result
    try:
        segments = []
        for byte in inet_pton(AF_INET6, network):
            segments += [ '%x' % (byte >> 4), '%x' % (byte & 0x0f) ]
        base_domain = 'ip6.arpa.'
        mask_divisor = 4
    except socket_error:
        segments = [ '%d' % byte for byte in inet_pton(AF_INET, network)]
        base_domain = 'in-addr.arpa.'
        mask_divisor = 8
    segments.reverse()
    n = mask // mask_divisor
    return ('.'.join(segments[-n:]).lower() + '.' + base_domain, 
                '%s/%s' % (network, mask))

def network_from_zone_name(name):
    """
    Form a network from a zone name.

    TODO: Finish and test this.  IPv6 mask code not complete!
    Partly inspired by dnspython dns.reverse.from_address()
    """
    if name.endswith('.in-addr.arpa.'):
        rev_str = name[:name.rfind('.in-addr.arpa.')]
        addr_list = rev_str.split('.')
        addr_list.reverse()
        mask = len(addr_list) * 8
        if mask not in (0, 8, 16, 24, 32):
            return None
        if len(addr_list):
            addr_str = '.'.join(addr_list)
            # Check address and make pretty
            try:
                 addr_str = inet_ntop(AF_INET, inet_pton(AF_INET, addr_str))
            except socket_error:
                return None
        else:
            addr_str = '0'
        return "%s/%s" % (addr_str, mask)
    elif name.endswith('.ip6.arpa.'):
        rev_str = name[:name.rfind('.ip6.arpa.')]
        addr_list = rev_str.split('.')
        addr_list.reverse()
        mask = len(addr_list) * 4
        if mask not in range(0, 65, 4):
            return None
        # Start here
        l = len(addr_list)
        bytes_2 = []
        i = 0 
        while i < l:
            bytes_2.append(''.join([x for x in addr_list[i:i+4]]))
            i += 4
        addr_str = ':'.join(bytes_2)
        # Check address and make pretty
        try:
            addr_str = inet_ntop(AF_INET6, inet_pton(AF_INET6, addr_str))
        except socket_error:
            return None
        return addr_str

    
    # Nothing can be done here!
    return None

def label_from_address(address):
    """
    Convert a network address to a reverse FQDN zone label

    Partly inspired by dnspython dns.reverse.from_address()
    """
    # Check routines also over in dms.zone_tool if this needs to be
    # changed
    try:
        segments = []
        for byte in inet_pton(AF_INET6, address):
            segments += [ '%x' % (byte >> 4), '%x' % (byte & 0x0f) ]
        base_domain = 'ip6.arpa.'
    except socket_error:
        segments = [ '%d' % byte for byte in inet_pton(AF_INET, address)]
        base_domain = 'in-addr.arpa.'
    segments.reverse()
    return '.'.join(segments).lower() + '.' + base_domain 

def address_from_label(rev_fqdn_label):
    """
    Convert an FQDN reverse label into a network address
    
    Partly inspired by dnspython dns.reverse.from_address()
    """
    if rev_fqdn_label.endswith('.in-addr.arpa.'):
        rev_str = rev_fqdn_label[:rev_fqdn_label.rfind('.in-addr.arpa.')]
        addr_list = rev_str.split('.')
        addr_list.reverse()
        addr_str = '.'.join(addr_list)
        # Check address and make pretty
        try:
             addr_str = inet_ntop(AF_INET, inet_pton(AF_INET, addr_str))
        except socket_error:
            return None
        return addr_str
    elif rev_fqdn_label.endswith('.ip6.arpa.'):
        rev_str = rev_fqdn_label[:rev_fqdn_label.rfind('.ip6.arpa.')]
        addr_list = rev_str.split('.')
        addr_list.reverse()
        l = len(addr_list)
        bytes_2 = []
        i = 0 
        while i < l:
            bytes_2.append(''.join([x for x in addr_list[i:i+4]]))
            i += 4
        addr_str = ':'.join(bytes_2)
        # Check address and make pretty
        try:
            addr_str = inet_ntop(AF_INET6, inet_pton(AF_INET6, addr_str))
        except socket_error:
            return None
        return addr_str

    # Nothing can be done here!
    return None

def new_soa_serial_no(current, name, db_soa_serial=None, candidate=None, 
        wrap_serial_next_time=False, date_stamp=None):
    """
    Calculate a new SOA serial number given the current one.
    
    This function shoul always provide a new serial that will
    enable moving back to the date based serial possible next update.
    New Serial must also be greater than current serial to ensure that
    any changes to SOA values are propagated.

    Setting warap_serial_next_time, and then doing another update will
    bring the SOA serial number back to YYYYMMDDnn conventional operations
    format.

    RFC 2316 Sec 3.4.2.2 says that SOA will not be updated at all unless
    new serial number is a positive increment on the current, as defined
    by modulo 2^32 arithmetic in RFC 1982 Section 3.
    """
    # As per RFC 1034 and 1035, SOA serial number is unsigned int32
    # [0 .. (2**32 -1)], hence modulo 32 arithmetic for SOA serial numbers
    # in RFC 1982.
    SERIAL_BITS = 32
    if date_stamp:
        date = date_stamp.timetuple()
    else:
        date = datetime.now().timetuple()
    new_date_serial = (00 + 100 * date.tm_mday + 10000 * date.tm_mon
                    + 1000000 * date.tm_year)

    # Maximum increment and addition formulae from RFC1982 Sec 3.1
    max_increment  = 2**(SERIAL_BITS -1) - 1
    max_update = (current + max_increment) % (2**SERIAL_BITS)

    if (wrap_serial_next_time):
        candidate = max_update
    elif (not candidate):
        candidate = new_date_serial
        # Something interesting, but you need to check that time between 
        # updates > refresh while doing this....  Could be troublesome.
        # check out chosen default candidate
        #if (candidate <= max_increment 
        #    and (current - candidate) 
        #        >= settings['soa_serial_wrap_threshold']):
        #    # Deal to any 'out of convention' serial numbers that sneak
        #    # in.  
        #    # FIXME: This will work until the 36th day of the 48th month
        #    # of the year 2147 or serial no 2147483647 ie max_increment...
        #    candidate = max_update

    # Two number line cases here, 1) max_update > current, and the wrap
    # case 2) max_update < current. Wrap occurs at (2^32 - 1).
    if (max_update > current):
        if (db_soa_serial and db_soa_serial > current 
                    and db_soa_serial < max_update):
            base = db_soa_serial
        elif (db_soa_serial and db_soa_serial == max_update):
            raise _SOASerialOcclusionError(name)
        else:
            base = current

        if (candidate > base and candidate <= max_update):
            update = candidate
        else:
            update = (base + 1) % (2**SERIAL_BITS)
    elif(max_update < current):
        if (db_soa_serial and (db_soa_serial < max_update 
                                or db_soa_serial > current)):
            base = db_soa_serial
        elif (db_soa_serial and db_soa_serial == max_update):
            raise _SOASerialOcclusionError(name)
        else:
            base = current

        if (candidate <= max_update or candidate > base):
            update = candidate
        else:
            update = (base + 1) % (2**SERIAL_BITS)
    else:
        # This is mathematically impossible for SERIAL_BITS = 32 
        # If it happens, this program has shifted to an alternate
        # reality of memory corruption
        raise _SOASerialArithmeticError(name)
    
    # SOA serial number can never be zero - RFC 2136 Sec 4.2
    if (update == 0):
        # If 0 is value of max_update, want to decrement so 
        # that update happens
        if (max_update == 0):
            update = (update - 1) % (2**SERIAL_BITS)
        else:
            update = (update + 1) % (2**SERIAL_BITS)

    if db_soa_serial and db_soa_serial == update:
        raise _SOASerialPublishedError(name)

    return update





#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Module for ZoneDataUtil mix in class for zone_engine

Split out so that changes can be seen more easily
"""


import re
from copy import copy

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import text

from magcode.core.globals_ import *
from magcode.core.database import sql_types
from dms.globals_ import *
from dms.exceptions import *
from dms.auto_ptr_util import check_auto_ptr_privilege
from dms.dns import RRTYPE_SOA
from dms.dns import RRTYPE_NS
from dms.dns import RRTYPE_A
from dms.dns import RRTYPE_AAAA
from dms.dns import RRTYPE_CNAME
from dms.dns import RRTYPE_MX
from dms.dns import RRTYPE_SRV
from dms.dns import RRTYPE_PTR
from dms.dns import RROP_DELETE
from dms.dns import RROP_UPDATE_RRTYPE
from dms.dns import RROP_ADD
from dms.dns import RROP_PTR_UPDATE
from dms.dns import RROP_PTR_UPDATE_FORCE
from dms.dns import validate_zi_hostname
from dms.dns import validate_zi_ttl
from dms.dns import is_inet_hostname
from dms.dns import label_from_address
from dms.dns import new_zone_soa_serial
import dms.database.zone_cfg as zone_cfg
from dms.database.zone_sm import exec_zonesm
from dms.database.zone_sm import ZoneSMDoRefresh
from dms.database.zone_sm import ZoneSM
from dms.database.reverse_network import ReverseNetwork
from dms.database.zone_instance import ZoneInstance
from dms.database.rr_comment import RRComment
from dms.database.resource_record import data_to_rr
from dms.database.resource_record import RR_PTR
from dms.database.resource_record import ResourceRecord
from dms.database.reference import find_reference
from dms.database.zi_update import ZiUpdate
from dms.database.update_group import new_update_group


class DataTools(object):
    """
    Container class for methods and runtime data for consistency 
    checking code 
    """

    def __init__(self, db_session, zone_sm, zi_cname_flag=False):
        """
        Initialise runtime data
        """
        self.db_session = db_session
        self.zone_sm = zone_sm
        self.name = zone_sm.name
        self.zi_cname_flag = zi_cname_flag
        self.zi_rr_data = {}
        self.auto_ptr_data = []
        self.apex_comment = None

    def check_rr_consistency(self, rrs, rr, rr_data, update_group):
        """
        Check that RR can be consistently added to zone
        """
        # Skip for any RROP_DELETE
        if update_group and rr.update_op and rr.update_op == RROP_DELETE:
            return

        if (not update_group or not rr.update_op 
                or rr.update_op != RROP_UPDATE_RRTYPE): 
            # Duplicate Record check
            if rr in rrs:
                raise DuplicateRecordInZone(self.name, rr_data)
            
            # Can't add another SOA if there is one there already
            if rr.type_ == RRTYPE_SOA:
                num_soa = len([r for r in rrs if r.type_ == RRTYPE_SOA])
                if num_soa:
                    raise ZoneAlreadyHasSOARecord(self.name, rr_data)

        # CNAME addition check
        if rr.type_ == RRTYPE_CNAME:
            self.zi_cname_flag = True

        # anti-CNAME addition check
        if self.zi_cname_flag:
            # Find any cnames with rr label and barf
            num_lbls = len([ r for r in rrs 
                            if (r.type_ == RRTYPE_CNAME 
                                and r.label == rr.label)])
            # Check that we are not updating an existing CNAME
            if (num_lbls and update_group and rr.update_op 
                    and rr.update_op == RROP_UPDATE_RRTYPE
                    and rr.type_ == RRTYPE_CNAME):
                num_lbls = 0
            if num_lbls:
                raise ZoneCNAMEExists(self.name, rr_data)

    def check_zi_consistency(self, rrs):
        """
        Check consistency of zone instance
        """
        # CNAME check
        rr_cnames = [r for r in rrs if r.type_ == RRTYPE_CNAME]
        for rr in rr_cnames:
            clash = len([ r for r in rrs 
                    if (r.label == rr.label and r.type_ != RRTYPE_CNAME)])
            if clash:
                raise ZoneCNAMELabelExists(self.name, self.zi_rr_data[str(rr)])
        # Check NS MX and SRV records point to actual A 
        # and AAAA records if they are in zone 
        # (Bind Option check-integrity)
        # NS
        rr_nss = [r for r in rrs if r.type_ == RRTYPE_NS 
                                                and r.label != '@']
        for rr in rr_nss:
            if not rr.rdata.endswith('.'):
                target_hosts = [r for r in rrs if r.label == rr.rdata]
                if not len(target_hosts):
                    raise ZoneCheckIntegrityNoGlue(self.name, 
                            self.zi_rr_data[str(rr)], rr.rdata)
        # MX
        rr_mxs = [r for r in rrs if r.type_ == RRTYPE_MX]
        for rr in rr_mxs:
            if not rr.rdata.endswith('.'):
                rdata = rr.rdata.split()
                target_hosts = [r for r in rrs if r.label == rdata[1]]
                if not len(target_hosts):
                    raise ZoneCheckIntegrityNoGlue(self.name, 
                            self.zi_rr_data[str(rr)], rdata[1])

        #SRV
        rr_srvs = [r for r in rrs if r.type_ == RRTYPE_SRV]
        for rr in rr_srvs:
            if not rr.rdata.endswith('.'):
                rdata = rr.rdata.split()

                target_hosts = [r for r in rrs if r.label == rdata[3]]
                if not len(target_hosts):
                    raise ZoneCheckIntegrityNoGlue(self.name, 
                            self.zi_rr_data[str(rr)], rdata[3])

        # If NS records are part of the zone, no point in doing
        # sanity checks as client will not be sending any SOAs
        if self.zone_sm.use_apex_ns:
            return
        # Check that zi has 1 SOA, and that its for the apex '@'
        rr_soas = [r for r in rrs if r.type_ == RRTYPE_SOA]
        if not rr_soas:
            raise ZoneHasNoSOARecord(self.name)
        if len(rr_soas) > 1:
            raise ZoneAlreadyHasSOARecord(self.name, 
                                self.zi_rr_data[str(rr_soas[1])])
        if rr_soas[0].label != '@':
            raise ZoneSOARecordNotAtApex(self.name, 
                                self.zi_rr_data[str(rr_soas[0])])
        # Check that apex has at least 1 NS record
        rr_nss = [r for r in rrs if r.type_ == RRTYPE_NS 
                and r.label == '@']
        if not rr_nss:
            raise ZoneHasNoNSRecord(self.name, 
                                        self.zi_rr_data[str(rr_soas[0])])

    def put_zi_rr_data(self, key, rr_data):
        """
        Store rr_data for later use
        """
        self.zi_rr_data[key] = rr_data
            
    def get_auto_ptr_data(self):
        """
        Return auto_ptr_data
        """
        return self.auto_ptr_data

    def handle_auto_ptr_data(self, rr, rr_data):
        """
        Handle auto reverse IP functionality.

        This is brief to quickly come up with a list of candidates
        that can be filtered for netblock reverse zone later on.
        """
        # We only look at IP address records
        if (rr.type_ != RRTYPE_A 
            and rr.type_ != RRTYPE_AAAA):
            return
       
        # We ignore DELETE update_ops, as algorithm will ignore that
        if (rr.update_op and rr.update_op == RROP_DELETE):
            return

        # Use the dnspython rewritten rdata to make sure that IPv6
        # addresses are uniquely written.
        hostname = rr.label + '.' + self.name if rr.label != '@' else self.name
        # Force reverse is once only, and not saved to DB, track_reverse is 
        # force reverse all the time
        force_reverse = False
        if rr_data.get('force_reverse'):
            force_reverse = True if rr_data['force_reverse'] else False
        if rr_data.get('track_reverse'):
            force_reverse = True if rr_data['track_reverse'] else force_reverse
        disable = False
        if rr_data.get('disable'):
            disable = True if rr_data['disable'] else False
        zone_ref = self.zone_sm.reference
        zone_ref_str = zone_ref.reference if zone_ref else None
        self.auto_ptr_data.append({ 'address': rr.rdata,
                                    'disable': disable,
                                    'force_reverse': force_reverse,
                                    'hostname': hostname,
                                    'reference': zone_ref_str})

    def check_reference_string(self, ref_str):
        """
        Check that the supplied reference string is complete
        """
        if not re.match(r'^[\-_a-zA-Z0-9.@]+$', ref_str):
            error_msg = "can only contain characters '-_a-zA-Z0-9.@'"
            raise ReferenceFormatError(ref_str, error_msg)
        if not re.match(r'^[0-9a-zA-Z][\-_a-zA-Z0-9.@]*$', ref_str):
            error_msg = "must start with 'a-zA-Z0-9'"
            raise ReferenceFormatError(ref_str, error_msg)
        if len(ref_str) > 1024:
            error_msg = "too long, must be <= 1024."
            raise ReferenceFormatError(ref_str, error_msg)

    def check_extra_data_privilege(self, rr_data, admin_privilege, 
                                    helpdesk_privilege):
        """
        Check privilege for use of extra data items to do with auto
        reverse IP setting and pyparsing error finformation
        """
        if (not admin_privilege):
            if (rr_data.get('lock_ptr')):
                raise AdminPrivilegeNeeded(self.name, rr_data,
                                        'lock_ptr')
                rr_data.pop('lock_ptr', None)
        if (not admin_privilege and not helpdesk_privilege):
            if rr_data.get('reference'):
                raise HelpdeskPrivilegeNeeded(self.name, rr_data,
                                        'reference')
                rr_data.pop('reference', None)

    def add_comment(self, top_comment, comment=None, tag=None, **kwargs):
        """
        Add a new comment or apex_comment
        """
        # Don't do anything unless 'comment' is supplied!
        if not comment and not top_comment:
            return None

        db_session = self.db_session
        # Deal with Apex comment - special, even set text to default 
        # if none!
        if (top_comment or tag == settings['apex_rr_tag']):
            if self.zone_sm.use_apex_ns:
                # If Apex done by global config, update routines 
                # will create an appropriate Apex comment
                return None
            if not comment:
                comment = settings['apex_comment_template'] % self.name
            tag = settings['apex_rr_tag']

        # Create a new comment
        rr_comment = RRComment(comment=comment, tag=tag)
        db_session.add(rr_comment)
        # Need to flush to get a new id from database
        db_session.flush()
        if (rr_comment.tag == settings['apex_rr_tag']):
            self.apex_comment = rr_comment
        return rr_comment.id_

    def get_apex_comment(self):
        """
        Return Apex Comment
        """
        return self.apex_comment
    
    def rr_data_create_comments(self, zi_data, zone_ttl, 
            creating_real_zi=True):
        """
        Common code for creating comments, and creating comment IDs
        """
        # Get comment IDs created and established.
        rr_group_data = zi_data.get('rr_groups')
        for rr_group in rr_group_data:
            rr_groups_index = rr_group_data.index(rr_group)
            top_comment = creating_real_zi and rr_groups_index == 0
            comment_group_id =  self.add_comment(top_comment, **rr_group)
            rr_group['comment_group_id'] = comment_group_id
            for rr_data in rr_group['rrs']:
                # get rr_groups_index and rrs_index for error handling
                rr_data['rrs_index'] = rr_group['rrs'].index(rr_data)
                rr_data['rr_groups_index'] = rr_groups_index
                # Handle comment IDs
                rr_data['comment_rr_id'] = self.add_comment(False, **rr_data)
                rr_data['comment_group_id'] = comment_group_id
                # Following needed to initialise dnspython RRs correctly
                rr_data['zone_ttl'] = zone_ttl
        self.rr_group_data = rr_group_data
        zi_data.pop('rr_groups', None)


    def add_rrs(self, rrs_func, add_rr_func, 
                admin_privilege, helpdesk_privilege,
                update_group=None):
        """
        Add RR to data base
        
        Note use of rrs_func so that list of rrs is always refreshed in
        function.  Can be supplied by using a no argument lambda function.
        This is so that in the case of a full ZI, rrs can be added to it,
        which is different to the case of incremental updates, where the 
        list of RRs is constructed, and the rrs just added directly to the
        resource records table.
        """
        db_session = self.db_session
        for rr_group in self.rr_group_data:
            for rr_data in rr_group['rrs']:
                # Remove unneeded keys from rr_data
                rr_data.pop('comment', None)
                rr_data.pop('zone_id', None)
                # Check privilege
                self.check_extra_data_privilege(rr_data, admin_privilege,
                        helpdesk_privilege)
                rr = data_to_rr(self.name, rr_data)
                self.check_rr_consistency(rrs_func(), rr, rr_data, update_group)
                # Store rr_data for zi consistency checks
                self.put_zi_rr_data(str(rr), rr_data)
                # Add rr to SQLAlchemy data structures
                db_session.add(rr)
                # Sort out RR reference part of the data structure
                rr_ref_str = rr_data.get('reference')
                if rr_ref_str: 
                    self.check_reference_string(rr_ref_str)
                    rr_ref = find_reference(db_session, rr_ref_str)
                    rr.ref_id = rr_ref.id_ if rr_ref else None
                    rr.reference = rr_ref
                # Sort out update_group if given
                if update_group:
                    update_group.update_ops.append(rr)
                add_rr_func(rr)
                self.handle_auto_ptr_data(rr, rr_data)

class PseudoZi(ZiUpdate):
    """
    Dummy ZI class so that ZiUpdate operations can do a trial run, so that 
    incremental updates can be consistency checked by zi checking code.
    """

    def __init__(self, db_session, zi):
        # make sure ZiUpdate runs in trial mode
        ZiUpdate.__init__(self, db_session=db_session, trial_run=True)
        # Copy rrs list so that changes do not trigger SQAlchemy
        self.rrs = []
        for rr in zi.rrs:
            rr_type = sql_types[type(rr).__name__]
            new_rr = rr_type(label=rr.label, domain=zi.zone.name, 
                    ttl=rr.ttl, zone_ttl=rr.zone_ttl,
                    rdata=rr.rdata, lock_ptr=rr.lock_ptr, disable=rr.disable,
                    track_reverse=rr.track_reverse)
            self.rrs.append(new_rr)


    def add_rr(self, rr):
        """
        Add RR to rrs list
        """
        self.rrs.append(rr)

    def remove_rr(self, rr):
        """
        Remove rr from rrs list
        """
        self.rrs.remove(rr)

class ZoneDataUtil(object):
    """
    Mix in class for ZoneEngine, containing _data_to_zi and _data_to_incr
    functions
    """

    def _data_to_zi(self, name, zi_data, change_by, normalize_ttls=False,
                admin_privilege=False, helpdesk_privilege=False):
        """
        Construct a new ZI, RRS and comments, from zone_data.
        """
            
        def set_missing_zi_data():
            """
            Set missing fields in supplied zi_data to prevent problems
            """
            # Set ZI Zone ttl if not already set
            if 'zone_ttl' not in zi_data:
                zi_data['zone_ttl'] = zone_ttl
            # Set other SOA values in zi_data from defaults 
            # if they are not there. soa_ttl can be None
            for field in ['soa_mname', 'soa_rname', 'soa_refresh', 'soa_retry', 
                    'soa_expire', 'soa_minimum']:
                if not zi_data.get(field):
                    zi_data[field] = zone_cfg.get_row_exc(db_session, field,
                                                        sg=zone_sm.sg)
            # We always update serial number on zone udpdate/publish
            # but it is nicer and probably less troublesome to replace 
            # an existing serial number that may be out there
            if not zi_data.get('soa_serial'):
                if zone_sm.soa_serial:
                    zi_data['soa_serial'] = zone_sm.soa_serial
                else:
                    # Obviously a new zone
                    zi_data['soa_serial'] = new_zone_soa_serial(db_session)

        def check_zi_data():
            """
            Check incoming zi_data attributes for correctness
            """
            for field in ['soa_mname', 'soa_rname']:
                validate_zi_hostname(name, field, zi_data[field])
            for field in ['soa_refresh', 'soa_retry', 'soa_expire',
                    'soa_minimum', 'soa_ttl', 'zone_ttl']:
                if field == 'soa_ttl' and not zi_data.get(field):
                    # SOA TTL can be None
                    continue
                validate_zi_ttl(name, field, zi_data[field])
            for field in ['soa_serial']:
                if field == 'soa_serial' and zi_data.get(field, None) == None:
                    # SOA serial can be None
                    continue
                # Check incoming data type of soa_serial
                if not isinstance(zi_data['soa_serial'], int):
                    raise SOASerialTypeError(name)
                if not ( 0 < zi_data['soa_serial'] <= (2**32-1)):
                    # RFC 2136 Section 4.2 AO serial cannot be zero
                    raise SOASerialRangeError(name)
            
        # Function start
        db_session = self.db_session
        # Get zone_sm to get zone ID etc
        zone_sm = self._get_zone_sm(name)
        zone_id = zone_sm.id_
        
        # initialise data and zone consistency checking
        data_tools = DataTools(db_session, zone_sm)
        
        # Sort out a candidate value for zone_ttl so that RRs can be created
        zone_ttl = zi_data.get('zone_ttl',
                zone_cfg.get_row_exc(db_session, 'zone_ttl', sg=zone_sm.sg))
        zone_ttl_supplied = 'zone_ttl' in zi_data

        # Create comments, and set up comment IDs, and stuff for handlng
        # RR Groups zi_data structures
        data_tools.rr_data_create_comments(zi_data, zone_ttl)

        # Deal with ZI data problems, and supply defaults if missing
        set_missing_zi_data()
        check_zi_data()
        # This constructor call sets attributes in zi as well!
        zi = ZoneInstance(change_by=change_by, **zi_data)
        db_session.add(zi)
        apex_comment = data_tools.get_apex_comment()
        if apex_comment:
            zi.add_apex_comment(apex_comment)
        # Get zi.id_ zi.zone_id from database
        db_session.flush()
        
        # Add RRs to zi
        # Note use of lambda so that list of rrs is always refreshed in
        # function
        data_tools.add_rrs(lambda :zi.rrs, zi.add_rr,
                admin_privilege, helpdesk_privilege)

        # tie zi into data_structures
        zone_sm.all_zis.append(zi)
        zi.zone = zone_sm
        db_session.flush()
        # Normalise TTLs here
        if normalize_ttls and zone_ttl_supplied:
            zi.normalize_ttls()
        # Update SOA and NS records - can't hurt to do it here
        # This also cleans out any incoming apex NS records if
        # client should not be setting them.
        zi.update_apex(db_session)
        # Update Zone TTLs for clean initialisation
        zi.update_zone_ttls()
        db_session.flush()
        # Check zone consistency. Do this here as Apex RRs need to be complete.
        data_tools.check_zi_consistency(zi.rrs)
        return zi, data_tools.get_auto_ptr_data()
    
    def _data_to_update(self, name, update_data, update_type, change_by,
            admin_privilege=False, helpdesk_privilege=False):
        """
        Construct an update group for a zone, from supplied RRS and comments.

        Functional equivalent of _data_to_zi() above, but for incremental 
        updates
        """
        # Function start
        db_session = self.db_session
        # Check that update_type is supplied
        if not update_type:
            raise UpdateTypeRequired(name)
        # Get zone_sm to get zone ID etc
        zone_sm = self._get_zone_sm(name)
        zone_id = zone_sm.id_

        # See if incremental updates are enabled for zone before queuing any
        if not zone_sm.inc_updates:
            raise IncrementalUpdatesDisabled(name)
        # Don't queue updates for a disabled zone
        if zone_sm.is_disabled():
            raise ZoneDisabled(name)
        # Privilege check for no apex zones - admin only 
        if not zone_sm.use_apex_ns and not admin_privilege:
            raise ZoneAdminPrivilegeNeeded(name)
       
        # Use candidate ZI as it always is available.  zi is published zi
        zi = self._get_zi(zone_sm.zi_candidate_id)
        if not zi:
            raise ZiNotFound(name, zone_sm.zi_candidate_id)

        # Get value of zone_ttl so that RRs can be created
        zone_ttl = zi.zone_ttl

        # Create RRs list from published ZI
        pzi = PseudoZi(db_session, zi)

        # initialise data and zone consistency checking
        zi_cname_flag = False
        if len([r for r in pzi.rrs if r.type_ == RRTYPE_CNAME]):
            zi_cname_flag = True
        data_tools = DataTools(db_session, zone_sm, zi_cname_flag)

        # Create comments, and set up comment IDs, and stuff for handlng
        # RR Groups zi_data structures
        data_tools.rr_data_create_comments(update_data, zone_ttl, 
                        creating_real_zi=False)
        try:
            # Create new update_group
            update_group = new_update_group(db_session, update_type, 
                                zone_sm, change_by)
        except IntegrityError as exc:
            raise UpdateTypeAlreadyQueued(name, update_type)

        # Add RRs to DB and operate on Pseudo ZI
        data_tools.add_rrs(lambda :pzi.rrs, pzi.trial_op_rr,
            admin_privilege, helpdesk_privilege, update_group=update_group)

        data_tools.check_zi_consistency(pzi.rrs)

        # Get all data out to DB, and ids etc established.
        db_session.flush()
        # Refresh zone to implement updates
        exec_zonesm(zone_sm, ZoneSMDoRefresh)

        # Return auto update info
        return data_tools.get_auto_ptr_data()

    def _queue_auto_ptr_data(self, auto_ptr_data):
        """
        Queue auto PTR data as incremental updates against respective reverse
        zones.
        """
        if not auto_ptr_data:
            return
        if not len(auto_ptr_data):
            return
        if not settings['auto_reverse']:
            return
        db_session = self.db_session
        # Create new update_group
        ug_dict = {}
        auto_ptr_privilege_flag = False
        for ptr_data in auto_ptr_data:
            # Ignore addresses we don't have reverse zone for
            query = db_session.query(ZoneSM)\
                    .join(ReverseNetwork)\
                    .filter(text(":address  <<= reverse_networks.network"))\
                    .params(address = ptr_data['address'])
            query = ZoneSM.query_is_not_deleted(query)
            query = ZoneSM.query_inc_updates(query)
            query = query.order_by(ReverseNetwork.network.desc())\
                    .limit(1)
            try:
                zone_sm = query.one()
            except NoResultFound:
                continue

            # Ignore invalid host names
            if not is_inet_hostname(ptr_data['hostname'], absolute=True,
                                    wildcard=False):
                log_error("Hostname '%s' is not a valid hostname." 
                        % ptr_data['hostname'])
                continue

            # Determine proposed update operation
            update_op = RROP_PTR_UPDATE_FORCE if ptr_data['force_reverse'] \
                            else RROP_PTR_UPDATE

            # Execute privilege checks ahead of time to save unnecessary churn
            # Better than needlessly going through whole rigamorole of 
            # incremental update processing later on for no effect
            #1 See if old PTR exists to retrieve any RR reference
            # Both following also used lower down when generating RR_PTR
            label = label_from_address(ptr_data['address'])
            rr_ref = find_reference(db_session, ptr_data['reference'],
                                        raise_exc=False)
            # query for old record - this generates one select
            # Optimization  - if check has previously suceeded, don't check
            # again as this is all checked further in
            if not auto_ptr_privilege_flag:
                qlabel= label[:label.rfind(zone_sm.name)-1]
                query = db_session.query(ResourceRecord)\
                    .filter(ResourceRecord.label == qlabel)\
                    .filter(ResourceRecord.zi_id == zone_sm.zi_candidate_id)\
                    .filter(ResourceRecord.disable == False)\
                    .filter(ResourceRecord.type_ == RRTYPE_PTR)
                old_rrs = query.all()
                old_rr = old_rrs[0] if len(old_rrs) else None
                
            # Check that we can proceed, only if check has not succeded yet
                if not check_auto_ptr_privilege(rr_ref, self.sectag, zone_sm,
                    old_rr):
                    if old_rr:
                        log_debug("Zone '%s' - can't replace '%s' PTR"
                            " as neither"
                            " sectags '%s' vs '%s'"
                            " references '%s' vs '%s'/'%s' (old PTR/rev zone)"
                            "match ,"
                            " or values not given."
                        % (zone_sm.name, old_rr.label,
                            self.sectag.sectag, settings['admin_sectag'],    
                            rr_ref, old_rr.reference, zone_sm.reference))
                    else:
                        log_debug("Zone '%s' - can't add '%s' PTR as neither"
                                " sectags '%s' vs '%s'"
                                " references '%s' vs '%s' (rev zone) match,"
                                " or values not given."
                            % (zone_sm.name, qlabel,
                                self.sectag.sectag, settings['admin_sectag'],    
                                rr_ref, zone_sm.reference))
                    continue
                auto_ptr_privilege_flag = True

            # Create a new update group if zone has not been seen before.
            try:
                update_group, zone_ttl = ug_dict.get(zone_sm)
            except (ValueError, TypeError):
                # Obtain reverse zone_ttl so PTR rrs can be created
                # Use candidate ZI as it always is available.
                # zi is published zi
                zi = self._get_zi(zone_sm.zi_candidate_id)
                if not zi:
                    log_error("Zone '%s': does not have candidate zi." 
                                % zone_sm.name)
                    continue
                zone_ttl = zi.zone_ttl
                update_group = new_update_group(db_session, None, 
                                        zone_sm, None, ptr_only=True, 
                                        sectag=self.sectag.sectag)
                ug_dict[zone_sm] = (update_group, zone_ttl)
            
            # Allocate RR_PTR update record
            rr = RR_PTR(label=label, zone_ttl=zone_ttl, 
                    rdata=ptr_data['hostname'], disable=ptr_data['disable'],
                    domain=zone_sm.name, update_op=update_op)
            rr.ref_id = rr_ref.id_ if rr_ref else None
            rr.reference = rr_ref

            # Chain on  RR_PTR update record
            update_group.update_ops.append(rr)

        # Flush everything to disk
        db_session.flush()
        # Issue zone refreshes to implement PTR changes
        for zone_sm in ug_dict:
            if zone_sm.is_disabled():
                continue
            exec_zonesm(zone_sm, ZoneSMDoRefresh)
        # Make sure everything is committed
        db_session.commit()



#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Zone instance copying using internal python data structures   Mix in class to modularise ZI classes.
"""


from magcode.core.globals_ import *
from magcode.core.database import *
from dms.dns import RRTYPE_A
from dms.dns import RRTYPE_AAAA


class ZiCopy(object):
    """
    Contains methods for ZI copying via internal structures
    """

    def copy(self, db_session, change_by=None):
        """
        Copy ZI

        First initialise base ZI, coppy  comment structure set up,
        then add records.
        """
        ZoneInstance = sql_types['ZoneInstance']
        RRComment = sql_types['RRComment']
        # Keep previous change_by if it is not being changed. This is useful
        # for auto PTR updates
        if not change_by:
            change_by = self.change_by
        new_zi = ZoneInstance(zone_id=self.zone_id, soa_serial=self.soa_serial,
                soa_mname=self.soa_mname, soa_rname=self.soa_rname,
                soa_refresh=self.soa_refresh, soa_retry=self.soa_retry,
                soa_expire=self.soa_expire, soa_minimum=self.soa_minimum,
                soa_ttl=self.soa_ttl, zone_ttl=self.zone_ttl,
                change_by=change_by)
        db_session.add(new_zi)
        new_zi.zone = self.zone
        
        # Establish Apex comment, which is a special RR_Groups comment
        # This dict establishes relN between new group comment and old one
        # by indexing the new comment against the old comments id
        rr_group_comments = {}
        if self.apex_comment:
            new_apex_comment = RRComment(comment=self.apex_comment.comment,
                                        tag=self.apex_comment.tag)
            db_session.add(new_apex_comment)
            new_zi.apex_comment = new_apex_comment
            rr_group_comments[self.apex_comment.id_] = new_apex_comment
        
        # Establish rest of RR_Groups comments
        for comment in self.rr_group_comments:
            if self.apex_comment and comment is self.apex_comment:
                # Apex comment is already done above here
                continue
            new_comment = RRComment(comment=comment.comment, tag=comment.tag)
            db_session.add(new_comment)
            rr_group_comments[comment.id_] = new_comment
        
        # For the sake of making code clearer, do same for RR_Comments as
        # for group comments
        rr_comments = {}
        for comment in self.rr_comments:
            new_comment = RRComment(comment=comment.comment, tag=comment.tag)
            db_session.add(new_comment)
            rr_comments[comment.id_] = new_comment

        # Walk zi RRs, and copy them as we go
        for rr in self.rrs:
            rr_type = sql_types[type(rr).__name__]
            new_rr = rr_type(label=rr.label, domain=self.zone.name, 
                    ttl=rr.ttl, zone_ttl=rr.zone_ttl,
                    rdata=rr.rdata, lock_ptr=rr.lock_ptr, disable=rr.disable,
                    track_reverse=rr.track_reverse)
            db_session.add(new_rr)
            new_zi.rrs.append(new_rr)
            if rr_group_comments.get(rr.comment_group_id):
                rr_group_comment = rr_group_comments[rr.comment_group_id]
                new_rr.group_comment = rr_group_comment
                # Uncomment if above is not 'taking'
                # rr_group_comment.rr_group.append(new_rr)
            if rr_comments.get(rr.comment_rr_id):
                rr_comment = rr_comments[rr.comment_rr_id]
                new_rr.rr_comment = rr_comment
                # Uncomment if above is not 'taking'
                # rr_comment.rr = new_rr
            if hasattr(rr, 'reference') and rr.reference:
                # Done this way as relationship is 'loose', 
                # SA relN is 'viewonly=True'
                new_rr.ref_id = rr.ref_id
        # Flush to DB to fill in record IDs
        db_session.flush()
        return new_zi

    def get_auto_ptr_data(self, zone_sm):
        """
        Return auto_ptr_data for the zi
        """
        auto_ptr_data = []
        zone_ref = zone_sm.reference
        zone_ref_str = zone_ref.reference if zone_ref else None
        for rr in self.rrs:
            if rr.type_ not in (RRTYPE_A, RRTYPE_AAAA):
                continue
            # Use the dnspython rewritten rdata to make sure that IPv6
            # addresses are uniquely written.
            hostname = rr.label + '.' + zone_sm.name \
                        if rr.label != '@' else zone_sm.name
            disable = rr.disable
            auto_ptr_data.append({ 'address': rr.rdata,
                                        'disable': disable,
                                        'force_reverse': False,
                                        'hostname': hostname,
                                        'reference': zone_ref_str})
        return auto_ptr_data




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Module to contain DMS Zone editing engine
"""

import json

import sqlalchemy.exc

from magcode.core import *
from magcode.core.database import sql_data
from magcode.core.wsgi.jsonrpc_server import *
from dms.zone_engine import ZoneEngine
from dms.exceptions import DMSError
from dms.exceptions import DBReadOnlyError
from dms.exceptions import ZoneHasNoSOARecord 


class DMSEngine(ZoneEngine):
    """
    Zone Editing Engine for use with the dms daemon.
    """

    def list_zone_helpdesk(self, names=None, reference=None, 
            include_deleted=False, toggle_deleted=False, include_disabled=True):
        """
        Help desk privilege list_zone()
        """
        return self._list_zone(names=names, reference=reference, 
                include_deleted=include_deleted, toggle_deleted=toggle_deleted,
                include_disabled=include_disabled)

    def update_zone_helpdesk(self, name, zi_data, login_id, 
                                edit_lock_token=None):
        """
        Update a zone with admin privilege
        """
        return self._update_zone(name, zi_data, login_id, edit_lock_token, 
                 helpdesk_privilege=True)
    
    def update_zone_text_helpdesk(self, name, zi_text, login_id, 
                            edit_lock_token=None):
        """
        Update a zone with admin privilege
        """
        return self._update_zone_text(name, zi_text, login_id, 
                            edit_lock_token, helpdesk_privilege=True)

    def update_zone(self, name, zi_data, login_id, edit_lock_token=None):
        """
        Update a zone with no privileges
        """
        return self._update_zone(name, zi_data, login_id, edit_lock_token)

    def update_zone_text(self, name, zi_text, login_id, edit_lock_token=None):
        """
        Update a zone with admin privilege
        """
        return self._update_zone_text(name, zi_text, login_id, edit_lock_token)

    def create_zone_helpdesk(self, name, login_id, zi_data=None, edit_lock=None,
            auto_dnssec=None, nsec3=None, inc_updates=None, 
            reference=None, sg_name=None):
        """
        Create a zone with admin privilege
        """
        return self._create_zone(name, zi_data, login_id, edit_lock=edit_lock,
                auto_dnssec=auto_dnssec, nsec3=nsec3, inc_updates=inc_updates, 
                reference=reference, sg_name=sg_name,
                use_apex_ns=None, helpdesk_privilege=True)
    

    def load_zone_helpdesk(self, name, login_id, zi_text, edit_lock=None,
            auto_dnssec=None, nsec3=None, inc_updates=None, 
            reference=None, sg_name=None):
        """
        Create a zone from a zone text blob with admin privilege
        """
        return self._load_zone(name, zi_text, login_id, edit_lock=edit_lock,
                auto_dnssec=auto_dnssec, nsec3=nsec3, inc_updates=inc_updates, 
                reference=reference, sg_name=sg_name,
                use_apex_ns=None, helpdesk_privilege=True)
    
    def load_zi_helpdesk(self, name, login_id, zi_text):
        """
        Load a zi text blob into a zone.  Help desk version
        """
        return self._load_zi(name, zi_text, login_id, helpdesk_privilege=True)

    def create_zone(self, name, reference, login_id, zi_data=None):
        """
        Create a zone with no privileges

        Note: inc_updates hard-wired here to True
        """
        return self._create_zone(name, zi_data, use_apex_ns=None,
                edit_lock=None, auto_dnssec=None, nsec3=None, inc_updates=True,
                reference=reference, login_id=login_id)
    
    def load_zone(self, name, reference, login_id, zi_text):
        """
        Load a zone from a zi_text blob. Customer version
        """
        return self._load_zone(name, zi_text, use_apex_ns=None,
                edit_lock=None, auto_dnssec=None, nsec3=None, inc_updates=True,
                reference=reference, login_id=login_id)

    def load_zi(self, name, login_id, zi_text):
        """
        Load a zi_text blob into a zone. Customer version
        """
        return self._load_zi(name, zi_text, login_id)

    def copy_zone_helpdesk(self, src_name, name, login_id, zi_id=None, 
                edit_lock=None, auto_dnssec=None,
                nsec3=None, inc_updates=None, reference=None, sg_name=None,
                sectags=None):
        """
        Copy a zone with helpdesk privilege
        """
        return self._create_zone(name, src_name=src_name, src_zi_id=zi_id,
                edit_lock=edit_lock, 
                auto_dnssec=auto_dnssec, nsec3=nsec3, inc_updates=inc_updates, 
                reference=reference, sg_name=sg_name, login_id=login_id, 
                zi_data=None,
                helpdesk_privilege=True)

    def copy_zone(self, src_name, name, login_id, zi_id=None):
        """
        Copy a zone
        """
        return self._create_zone(name, src_name=src_name, src_zi_id=zi_id, 
                                login_id=login_id, zi_data=None)

    def delete_zone_helpdesk(self, name):
        """
        Delete a zone helpdesk front end
        """
        self._delete_zone(name, force=True) 

    def set_zone_helpdesk(self, name, **kwargs): 
        for arg in kwargs:
            if arg not in ('edit_lock', 'auto_dnssec',):
                raise InvalidParamsJsonRpcError("Argument '%s' not supported."
                                                    % arg)
        return self._set_zone(name, **kwargs)
    
    def update_rrs(self, name, update_data, update_type, login_id):
        """
        Incremental updates, normal customer api privilege
        """
        return self._update_rrs(name, update_data, update_type, login_id)

    def update_rrs_helpdesk(self, name, update_data, update_type, login_id):
        """
        Incremental updates, help desk privilege
        """
        return self._update_rrs(name, update_data, update_type, login_id,
                                helpdesk_privilege=True)

    # Test code for exception raising
    #def list_zone(self, *args):
    #   raise ZoneHasNoSOARecord(args[0])

class BaseJsonRpcContainer(object):
    
    """
    Implements mapping between JSON RPC method names, and the appropriate
    engine methods.  Each application should descend from this class, and 
    define methods, in a class called 'JSONRpcCaller'.  
    The application function will create an instance of that class, 
    """
    def __init__(self, sectag, time_format=None):
        self._engine = DMSEngine(time_format=time_format, sectag_label=sectag)

    def _exc_rollback(self):
        """
        Cleanup db_session if there is an exception!
        """
        self._engine.rollback()

class DmsJsonRpcServer(object):
    """
    DMS JSON RPC Server class

    Creates a callable object that has the RPC call security container class
    and sectag as attributes.  This allows per request initialisation of
    SQL Alchemy session etc.
    """
    def __init__(self, rpc_container_class, sectag, time_format=None):
        self.rpc_container_class = rpc_container_class
        self.sectag = sectag
        self.time_format = time_format

    def __call__(self, environ, start_response, requests):
            
        # Initialise DB and engine object
        rpc_container = self.rpc_container_class(
                                                time_format=self.time_format, 
                                                sectag=self.sectag)
        
        # Process requests
        response = []
        for request in requests:
            # Process request
            call_id = request.get('id')
            if not call_id:
                # Skip any notifications at the moment
                continue
            params = request.get('params')
            if not hasattr(rpc_container, request['method']):
                response.append({'jsonrpc': '2.0', 'id':call_id,
                        'error': { 'code': JSONRPC_METHOD_NOT_FOUND,
                        'message': jsonrpc_errors[JSONRPC_METHOD_NOT_FOUND]}})
            # Double nested EXC so that standard exception processing
            # happens for PostgresQL in Read Only hot-standby.  This is lowest
            # common point where this can be trapped properly.
            try:
                try:
                    # Sort out params  - needs to be feed in correctly as
                    # python *args or **kwargs depending wether it is a JSON
                    # array or JSON object
                    if isinstance(params, list):
                        result = getattr(rpc_container,
                                    request['method'])(*params)
                    elif isinstance(params, dict):
                        result = getattr(rpc_container, 
                                    request['method'])(**params)
                    else:
                        result = getattr(rpc_container, request['method'])()
                    response.append({'id': call_id, 'result': result,
                                    'jsonrpc': '2.0'})
                except sqlalchemy.exc.InternalError as exc:
                    raise DBReadOnlyError(str(exc))
            except BaseJsonRpcError as exc:
                rpc_container._exc_rollback() 
                data = exc.data
                data.update({'exception_message': str_exc(exc),
                                'exception_type': str_exc_type(exc)})
                if jsonrpc_error_stack_trace():
                    data['stack_trace'] = format_exc()
                response.append({'jsonrpc': '2.0', 'id':call_id,
                        'error': { 'code': exc.jsonrpc_error,
                        'message': str(exc),
                        'data': data}})
            except (TypeError,AttributeError) as exc:
                rpc_container._exc_rollback() 
                data = {'exception_message': str_exc(exc),}
                if jsonrpc_error_stack_trace():
                    data['stack_trace'] = format_exc() 
                response.append({'jsonrpc': '2.0', 'id':call_id,
                        'error': { 'code': JSONRPC_INVALID_PARAMS,
                        'message': jsonrpc_errors[JSONRPC_INVALID_PARAMS],
                        'data': data}})
        return response




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Access zone_config table containing default values for zone initialisation, 
and Apex NS server names
"""


from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql import and_
from sqlalchemy.sql import or_

from magcode.core.database import *
from dms.database.sg_utility import find_sg_byname
from dms.exceptions import ZoneCfgItemNotFound
from dms.exceptions import NoSgFound


@saregister
class ZoneCfg(object):
    """
    Zone Config row label value object
    """
    _table = 'zone_cfg'

    def __init__(self, key, value, sg_id=None):
        self.key = key
        self.value = value
        self.sg_id = sg_id

    def to_engine(self, time_format=None):
        """
        Dict for JSON serialized output
        """
        if (hasattr(self, 'sg') and self.sg):
            sg_name = self.sg.name
        else:
            sg_name = None
        return {'key': self.key, 'sg_name': sg_name, 'value': self.value}
    
    to_engine_brief = to_engine

def get_row(db_session, key, sg=None, sg_name=None, raise_exc=False):
    """
    Return the first value found for a key

    if key not found in sg, return value where sg_id is None
    """
    result = None
    if sg_name:
        sg = find_sg_byname(db_session, sg_name)
    if sg:
        stuff = [x for x in sg.zone_cfg_entries if x.key == key]
        if stuff:
            result = stuff[0].value
            return result
    try:
        stuff = db_session.query(ZoneCfg)\
                    .filter(and_(ZoneCfg.key == key, 
                        ZoneCfg.sg_id == None)).all()
        result = stuff[0].value
    except IndexError:
        result = None
    if raise_exc and not result:
        raise ZoneCfgItemNotFound(key)
    return result

def get_row_exc(db_session, key, sg=None, sg_name=None):
    """
    Return the first value found for a key

    Raises Exception suitable for JSON RPC
    """
    return get_row(db_session, key, sg=sg, sg_name=sg_name, raise_exc=True)

def get_rows(db_session, key, sg=None, sg_name=None, raise_exc=False):
    """
    Return all the values for a key as a list
    """
    result = []
    if sg_name:
        sg = find_sg_byname(db_session, sg_name)
    if sg:
        stuff = [x for x in sg.zone_cfg_entries if x.key == key]
        result = [x.value for x in stuff]
        if result:
            return result
    try:
        stuff = db_session.query(ZoneCfg)\
                    .filter(and_(ZoneCfg.key == key, 
                                ZoneCfg.sg_id == None)).all()
        result =  [x.value for x in stuff]
    except NoResultFound:
        result = []
    if raise_exc and not result:
        raise ZoneCfgItemNotFound(key)
    return result

def get_rows_exc(db_session, key, sg=None, sg_name=None):
    """
    Return all the values found for a key

    Raises Exception suitable for JSON RPC
    """
    return get_rows(db_session, key, sg=sg, sg_name=sg_name, raise_exc=True)

def set_row(db_session, key, value, sg=None, sg_name=None):
    """
    Set one row to a given value

    This is always called from command line or wsgi configuration code.
    """
    if sg_name:
        sg = find_sg_byname(db_session, sg_name)
        if not sg:
            raise NoSgFound(sg_name)
    if sg:
        stuff = [x for x in sg.zone_cfg_entries if x.key == key]
        if stuff:
            zone_cfg = stuff[0]
            zone_cfg.value = value
            db_session.flush()
            return
        zone_cfg = ZoneCfg(key, value)
        db_session.add(zone_cfg)
        sg.zone_cfg_entries.append(zone_cfg)
        db_session.flush()
        return

    # We have reached the part which processes the case of no SG being given
    # to function
    try:
        stuff = db_session.query(ZoneCfg)\
                    .filter(and_(ZoneCfg.key == key,
                                ZoneCfg.sg_id == None)).one()
        zone_cfg = stuff
        zone_cfg.value = value
    except NoResultFound:
        zone_cfg = ZoneCfg(key, value)
        db_session.add(zone_cfg)
    finally:
        db_session.flush()

def set_rows(db_session, key, values, sg=None, sg_name=None):
    """
    Set a whole key type to a list of values

    This is always called from command line or wsgi configuration code.
    """
    if sg_name:
        sg = find_sg_byname(db_session, sg_name)
        if not sg:
            raise NoSgFound(sg_name)
    sg_id = sg.id_ if sg else None
    try:
        # Easiest to delete and recreate
        stuff = db_session.query(ZoneCfg)\
                    .filter(and_(ZoneCfg.key == key,
                            ZoneCfg.sg_id == sg_id)).all()
        for zone_cfg in stuff:
            db_session.delete(zone_cfg)
    except NoResultFound:
        pass

    for value in values:
        zone_cfg = ZoneCfg(key, value, sg_id)
        db_session.add(zone_cfg)
    # Flush should reconstruct sg zone_cfg_entries lists?
    # Any how, this function is called in configuration code, and this is
    # the end of the query group for that.  ie - data will commited on function
    # call return
    db_session.flush()


    




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Reverse Net table
"""


from sqlalchemy.orm import relationship

from magcode.core.database import *


@saregister
class ReverseNetwork(object):
    """
    Reverse Network Object to tie in to a zone
    """
    _table = 'reverse_networks'
    
    def __init__(self, network):
        """
        Initialise a reverse network object
        """
        self.network = network

def new_reverse_network(db_session, network):
    """
    Create a new reverse network
    """
    reverse_network = ReverseNetwork(network)
    db_session.add(reverse_network)
    db_session.flush()
    return reverse_network




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Module to contain Text Editor Zone editing engine
"""


from datetime import timedelta
import tempfile
import io
import os
import re
import errno
import pwd
import grp
from os.path import basename
from subprocess import check_call
from subprocess import check_output
from subprocess import CalledProcessError
from subprocess import STDOUT
from base64 import b64encode

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.sql import or_
from sqlalchemy.sql import and_
from sqlalchemy.sql import func
from sqlalchemy.sql import not_
from sqlalchemy.sql import select
from sqlalchemy.sql import delete
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import desc

from magcode.core.globals_ import *
from dms.globals_ import *
from magcode.core.utility import get_numeric_setting
from magcode.core.database import sql_types
from magcode.core.database import sql_data
from magcode.core.database.event import Event
from magcode.core.database.event import ESTATE_NEW
from magcode.core.database.event import ESTATE_RETRY
from magcode.core.database.event import ESTATE_FAILURE
from magcode.core.database.event import ESTATE_SUCCESS
from magcode.core.database.event import cancel_event
from magcode.core.database.event import event_processed_states
from dms.exceptions import *
from dms.zone_engine import ZoneEngine
from dms.database import zone_cfg
from dms.database.zone_cfg import ZoneCfg
from dms.database.zone_sm import ZoneSM
from dms.database.zone_sm import ZSTATE_DELETED
from dms.database.zone_sm import ZSTATE_DISABLED
from dms.database.zone_sm import ZSTATE_PUBLISHED
from dms.database.zone_sm import exec_zonesm
from dms.database.zone_sm import ZoneSMDoRefresh
from dms.database.zone_sm import ZoneSMNukeStart
from dms.database.zone_sm import ZoneSMDoDestroy
from dms.database.zone_sm import ZoneSMDoReset
from dms.database.server_group import ServerGroup
from dms.exceptions import ZoneNotFound
from dms.exceptions import ZiNotFound
from dms.exceptions import NoZonesFound
from dms.dns import validate_zi_ttl
from dms.dns import validate_zi_hostname
from dms.database.zone_sectag import new_sectag
from dms.database.zone_sectag import del_sectag
from dms.database.master_sm import reset_master_sm
from dms.database.master_sm import reconfig_all
from dms.database.master_sm import reconfig_sg
from dms.database.master_sm import reconfig_master
from dms.database.master_sm import get_master_sm
from dms.database.master_sm import get_mastersm_replica_sg
from dms.database.zone_instance import ZoneInstance
from dms.database.sg_utility import find_sg_byname
from dms.database.sg_utility import find_sg_byid
from dms.database.sg_utility import list_all_sgs
from dms.database.sg_utility import new_sg
from dms.database.sg_utility import set_sg_config
from dms.database.sg_utility import set_sg_master_address
from dms.database.sg_utility import set_sg_master_alt_address
from dms.database.sg_utility import set_sg_replica_sg
from dms.database.sg_utility import del_sg
from dms.database.sg_utility import rename_sg
from dms.database.server_sm import ServerSM
from dms.database.server_sm import SSTYPE_BIND9
from dms.database.server_sm import SSTYPE_NSD3
from dms.database.server_sm import server_types
from dms.database.server_sm import find_server_byname
from dms.database.server_sm import find_server_byaddress
from dms.database.server_sm import new_server
from dms.database.server_sm import del_server
from dms.database.server_sm import set_server
from dms.database.server_sm import rename_server
from dms.database.server_sm import set_server_ssh_address
from dms.database.server_sm import move_server_sg
from dms.database.server_sm import exec_server_sm
from dms.database.server_sm import ServerSMEnable
from dms.database.server_sm import ServerSMDisable
from dms.database.server_sm import ServerSMReset
from dms.database.syslog_msg import SyslogMsg


config_keys = ['soa_mname', 'soa_rname', 'soa_refresh', 'soa_retry',
            'soa_expire', 'soa_minimum', 'soa_ttl', 'zone_ttl', 'use_apex_ns',
            'edit_lock', 'auto_dnssec', 'default_sg', 'default_ref',
            'default_stype', 'zi_max_age', 'zi_max_num', 'zone_del_pare_age',
            'zone_del_age', 'event_max_age', 'syslog_max_age',
            'nsec3', 'inc_updates']

tsig_key_algorithms = ('hmac-md5', 'hmac-sha1', 'hmac-sha224', 'hmac-sha256',
        'hmac-sha384', 'hmac-512')

# 2 simple exceptions for restore_named_db to enable MasterSM.write_named_conf 
# and ZoneSM.write_zone_file to be called.
class ZoneFileWriteInternalError(Exception):
    pass
class NamedConfWriteInternalError(Exception):
    pass

class CmdLineEngine(ZoneEngine):
    """
    Zone Editing Engine for use with the command line and a text editor.
    """
    
    def __init__(self, sectag_label=None):
        super().__init__(time_format="%a %b %e %H:%M:%S %Y", 
                sectag_label=sectag_label)
    
    def show_zone_full(self, name, zi_id=None):
        """
        Given a zone name, return all the values stored in its ZoneSM
        record, current zi, all RRs, and comments
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        return self._show_zone(zone_sm, zi_id, all_rrs=True)

    def show_zone_byid_full(self, zone_id, zi_id=None):
        """
        Given a zone id, return all the values stored in its ZoneSM
        record, current zi, RRs, and comments
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        zone_sm = self._get_zone_sm_byid(zone_id)
        return self._show_zone(zone_sm, zi_id, all_rrs=True)

    def get_config_default(self, config_key):
        """
        Get the default value for a configuration key
        """
        self.refresh_db_session()
        return zone_cfg.get_row_exc(self.db_session, config_key)

    def set_config(self, config_key, value, sg_name=None):
        """
        Set a configuration item in the zone_cfg table
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        if config_key in ['soa_mname', 'soa_rname']:
            validate_zi_hostname(None, config_key, value)
        if config_key in ['soa_refresh', 'soa_retry', 'soa_expire', 'soa_ttl',
                'zone_ttl']:
            validate_zi_ttl(None, config_key, value)
        if config_key in ['soa_mname',]:
            if not sg_name:
               raise SgNameRequired(config_key)
        else:
            sg_name = None
        zone_cfg.set_row(self.db_session, config_key, value, sg_name=sg_name)
        self._finish_op()
        return {}

    def show_config(self):
        """
        Display all the configuration keys
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        db_session = self.db_session
        result = db_session.query(ZoneCfg)\
                    .filter(ZoneCfg.key.in_(config_keys)).all()
        result_list = []
        for zone_cfg in result:
            result_list.append(zone_cfg.to_engine())
        return result_list

    def show_apex_ns(self, sg_name=None):
        """
        Display the apex NS server settings
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        sg = find_sg_byname(self.db_session, sg_name)
        if not sg:
            raise NoSgFound(sg_name)
        result = zone_cfg.get_rows_exc(self.db_session, 
                                        settings['apex_ns_key'],
                                        sg_name=sg_name)
        return result
    
    def set_apex_ns(self, ns_servers, sg_name):
        """
        Set the apex NS server settings
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        # Strip blank lines
        ns_servers = [ns for ns in ns_servers if ns]
        # Fix up people not giving FQDN including root zone!
        ns_servers = [ (ns + '.' if not ns.endswith('.') else ns) 
                            for ns in ns_servers]
        result = zone_cfg.set_rows(self.db_session, settings['apex_ns_key'],
                                    ns_servers, sg_name=sg_name)
        self._finish_op()
        return result

    def nuke_zones(self, *names, include_deleted=False, toggle_deleted=False, 
            sg_name=None, reference=None):
        """
        Destroy multiple zones.  Multiple names may be given.  Wildcards
        can be used for partial matches.

        This is mainly a command for testing, or cleaning up after a large
        batch zone load goes awry.

        Zones being nuked have their deleted_start set to 1/1/1970, midnight.
        This means they will be immediately reaped by the next vacuum_zones
        command.
        """
        # Deal with SA auto-BEGIN - want fresh transaction to see fresh data
        self._begin_op()
        if not names:
            # No arguments
            self._finish_op()
            raise NoZonesFound('')
       
        db_session = self.db_session
        db_query_slice = get_numeric_setting('db_query_slice', int)
        # We were given some arguments
        zones = []
        # We keep domains and labels in database lowercase
        names = [x.lower() for x in names]
        name_pattern = ' '.join(names)
        names = [x.replace('*', '%') for x in names]
        names = [x.replace('?', '_') for x in names]
        for name in names:
            if not name.endswith('.') and not name.endswith('%'):
                name += '.'
            query = db_session.query(ZoneSM)\
                    .filter(ZoneSM.name.like(name))
            # Don't delete any reverse zones with this command
            query = query.filter(not_(ZoneSM.name.like('%.in-addr.arpa.')))\
                    .filter(not_(ZoneSM.name.like('%.ip6.arpa.')))
            if reference:
                query = query.join(Reference)\
                        .filter(Reference.reference.ilike(reference))
            if sg_name and self.sectag.sectag == settings['admin_sectag']:
                if sg_name not in list_all_sgs(self.db_session):
                    raise NoSgFound(sg_name)
                query = query.join(ServerGroup, ZoneSM.sg_id 
                                    == ServerGroup.id_)\
                        .filter(ServerGroup.name == sg_name)
            if include_deleted:
                pass
            elif toggle_deleted:
                query = query.filter(ZoneSM.state == ZSTATE_DELETED)
            else:
                query = query.filter(ZoneSM.state != ZSTATE_DELETED)
            query = query.yield_per(db_query_slice)
            # The following gives less RAM piggery even though it is slower
            for z in query:
                zones.append(z)
        # Take note of security tags
        if self.sectag.sectag != settings['admin_sectag']:
            zones = [x for x in zones if self.sectag in x.sectags]
        if not zones:
            if len(name_pattern) > 240:
                name_pattern = '* - %s names' % len(names)
            raise NoZonesFound(name_pattern)
        # Mark them all as deleted.
        for zone in zones:
            exec_zonesm(zone, ZoneSMNukeStart)
        self._finish_op()

    def create_sectag(self, sectag_label):
        """
        Create a new security tag
        """
        self._begin_op()
        new_sectag(self.db_session, sectag_label)
        self._finish_op()

    def delete_sectag(self, sectag_label):
        """
        Delete a security tag
        """
        self._begin_op()
        del_sectag(self.db_session, sectag_label)
        self._finish_op()
    
    def reset_mastersm(self):
        """
        Reset the Configuration state machine 
        """
        self._begin_op()
        reset_master_sm(self.db_session)
        self._finish_op()
    
    def _find_sg_byid(self, sg_id):
        """
        Given an sg_id, return the server group
        """
        db_session = self.db_session
        return find_sg_byid(db_session, sg_id, raise_exc=True)

    def create_sg(self, sg_name, config_dir=None, address=None, 
            alt_address=None, replica_sg=False):
        """
        Create a new SG
        """
        self._begin_op()
        new_sg(self.db_session, sg_name, config_dir, address, alt_address,
                replica_sg)
        self._finish_op()

    def rename_sg(self, sg_name, new_sg_name):
        """
        Rename an SG 
        """
        self._begin_op()
        rename_sg(self.db_session, sg_name, new_sg_name)
        self._finish_op()
        
    def set_sg_config(self, sg_name, config_dir=None):
        """
        Set the SG config dir
        """
        self._begin_op()
        set_sg_config(self.db_session, sg_name, config_dir)
        self._finish_op()

    def set_sg_master_address(self, sg_name, address=None):
        """
        Set the SG master server address
        """
        self._begin_op()
        set_sg_master_address(self.db_session, sg_name, address)
        self._finish_op()

    def set_sg_master_alt_address(self, sg_name, alt_address=None):
        """
        Set the alternate SG master server address
        """
        self._begin_op()
        set_sg_master_alt_address(self.db_session, sg_name, alt_address)
        self._finish_op()

    def set_sg_replica_sg(self, sg_name):
        """
        Set the replica_sg flag on an SG
        """
        self._begin_op()
        set_sg_replica_sg(self.db_session, sg_name)
        self._finish_op()

    def delete_sg(self, sg_name):
        """
        Delete an SG
        """
        self._begin_op()
        del_sg(self.db_session, sg_name)
        self._finish_op()

    def reconfig_all(self):
        """
        Reconfigure all servers
        """
        self._begin_op()
        reconfig_all(self.db_session)
        self._finish_op()

    def reconfig_sg(self, sg_name):
        """
        Reconfigure an SGs servers
        """
        self._begin_op()
        sg = self._find_sg_byname(sg_name)
        reconfig_sg(self.db_session, sg.id_, sg.name)
        self._finish_op()

    def reconfig_replica_sg(self):
        """
        Reconfigure replica SG.  
        
        This forces an rsync of DNSSEC key material and Zone data to DR
        replica servers.
        """
        self._begin_op()
        db_session = self.db_session
        replica_sg = get_mastersm_replica_sg(db_session)
        if not replica_sg:
            # If no replica_sg, return with no error
            return
        reconfig_sg(db_session, replica_sg.id_, replica_sg.name)
        self._finish_op()

    def reconfig_master(self):
        """
        Reconfigure the master DNS server
        """
        self._begin_op()
        reconfig_master(self.db_session)
        self._finish_op()

    def refresh_all(self):
        """
        Reconfigure all zones
        """
        self._begin_op()
        db_session = self.db_session
        id_query = db_session.query(ZoneSM.id_, ZoneSM.name)
        id_query = ZoneSM.query_is_configured(id_query)
        id_result = id_query.all()
        for zone_id, zone_name in id_result:
            try:
                zone_sm = db_session.query(ZoneSM)\
                        .filter(ZoneSM.id_ == zone_id).one()
            except NoResultFound:
                raise ZoneNotFoundByZoneId(zone_id)
            exec_zonesm(zone_sm, ZoneSMDoRefresh)
        self._finish_op()

    def refresh_sg(self, sg_name):
        """
        Refresh all zones on an SG
        """
        self._begin_op()
        db_session = self.db_session
        sg = self._find_sg_byname(sg_name)
        id_query = db_session.query(ZoneSM.id_, ZoneSM.name)\
                .filter(ZoneSM.sg_id == sg.id_) 
        id_query = ZoneSM.query_is_configured(id_query)
        id_result = id_query.all()
        for zone_id, zone_name in id_result:
            try:
                zone_sm = db_session.query(ZoneSM)\
                        .filter(ZoneSM.id_ == zone_id).one()
            except NoResultFound:
                raise ZoneNotFoundByZoneId(zone_id)
            exec_zonesm(zone_sm, ZoneSMDoRefresh)
        self._finish_op()

    def create_zone_batch(self, name, login_id, zi_data=None, 
                use_apex_ns=None, edit_lock=None, auto_dnssec=None,
                nsec3=None, inc_updates=None, reference=None, 
                sg_name=None, sectags=None):

        """
        Create a zone with admin privilege when doing a batch load
        """
        return self._create_zone(name, zi_data, login_id, use_apex_ns,
                edit_lock, auto_dnssec, nsec3, inc_updates, 
                reference, sg_name, sectags, admin_privilege=True,
                batch_load=True)

    def create_zi_zone_admin(self, zi_id, name, login_id,
            use_apex_ns=None, edit_lock=None, auto_dnssec=None,
            nsec3=None, inc_updates=None, reference=None, sg_name=None,
            sectags=None):
        """
        Create a zone with admin privilege
        """
        return self._create_zone(name, src_zi_id=zi_id,
                use_apex_ns=use_apex_ns, edit_lock=edit_lock, 
                auto_dnssec=auto_dnssec, nsec3=nsec3, inc_updates=inc_updates, 
                reference=reference, sg_name=sg_name, sectags=sectags,
                login_id=login_id, zi_data=None, admin_privilege=True)

    def vacuum_event_queue(self, age_days=None):
        """
        Destroy events processed more than age_days ago
        """
        self._begin_op()
        db_session = self.db_session
        if age_days is None:
            age_days = float(zone_cfg.get_row_exc(db_session,
                                key='event_max_age'))
        age_days = timedelta(days=age_days)
        count = 0
        # Do a straight SQL DELETE first to speed things along
        # Count events to be deleted
        event_table = sql_data['tables'][Event]
        where_stmt = and_(Event.state.in_(event_processed_states),
                            Event.processed != None,
                            (func.now() - Event.processed) > age_days)
        count_select = select([func.count(event_table.c.get('id'))],
                                where_stmt)
        result = db_session.execute(count_select).fetchall()
        count += result[0][0]
        db_session.execute(event_table.delete().where(where_stmt))
                    
        result = {'num_deleted': count}
        self._finish_op()
        return result

    def vacuum_zones(self, age_days=None):
        """
        Destroy zones older than age_days
        """
        self._begin_op()
        db_session = self.db_session
        db_query_slice = get_numeric_setting('db_query_slice', int)
        age_days_from_config = float(zone_cfg.get_row_exc(db_session,
                                    key='zone_del_age'))
        if age_days_from_config <= 0 and age_days is None:
            age_days = get_numeric_setting('zone_del_off_age', float)
        elif age_days is None:
            age_days = age_days_from_config
        age_days = timedelta(days=age_days)
        count = 0
        # Clear old and nuked zones one by one
        id_query = db_session.query(ZoneSM.id_)\
                .filter(ZoneSM.state == ZSTATE_DELETED)\
                .filter(or_(ZoneSM.deleted_start == None,
                            (func.now() - ZoneSM.deleted_start) > age_days))\
                .filter(ZoneSM.zone_files == False)\
                .yield_per(db_query_slice)
        id_results = []
        for zone_id, in id_query:
            id_results.append(zone_id)
        for zone_id in id_results:
            try:
                zone_sm = db_session.query(ZoneSM)\
                        .filter(ZoneSM.id_ == zone_id).one()
            except NoResultFound:
                continue
            if zone_sm.state != ZSTATE_DELETED:
                # Skip this if a customer has undeleted zone in the mean time..
                continue
            db_session.delete(zone_sm)
            db_session.commit()
            count += 1
                    
        # Finally do zone_sm destroy operation to 
        query = db_session.query(ZoneSM)\
                .filter(ZoneSM.state == ZSTATE_DELETED)\
                .filter(or_(ZoneSM.deleted_start == None,
                    (func.now() - ZoneSM.deleted_start) > age_days))
        for zone_sm in query:
            if zone_sm.state != ZSTATE_DELETED:
                # Skip this if a customer has undeleted zone in the mean time..
                continue
            try:
                exec_zonesm(zone_sm, ZoneSMDoDestroy)
            except ZoneSmFailure:
                continue
            count += 1
        result = {'num_deleted': count}
        self._finish_op()
        return result


    def vacuum_zis(self, age_days=None, zi_max_num=None):
        """
        Age ZIs according to age_days and zi_max_num
        """
        self._begin_op()
        db_session = self.db_session
        db_query_slice = get_numeric_setting('db_query_slice', int)
        if age_days is None:
            age_days = float(zone_cfg.get_row_exc(db_session,
                                key='zi_max_age'))
        age_days = timedelta(days=age_days)
        if zi_max_num is None:
            zi_max_num = int(zone_cfg.get_row_exc(db_session, 
                    key='zi_max_num'))
        stmt = db_session.query(ZoneInstance.zone_id,
                func.count(ZoneInstance.id_).label('zi_count'))\
                        .group_by(ZoneInstance.zone_id).subquery()
        zone_sm_query = db_session.query(ZoneSM)\
                .filter(ZoneSM.state != ZSTATE_DELETED)\
                .outerjoin(stmt, ZoneSM.id_ == stmt.c.zone_id)\
                .filter(stmt.c.zi_count > zi_max_num)\
                .yield_per(db_query_slice)
        count = 0
        for zone_sm in zone_sm_query:
            zi_keep = db_session.query(ZoneInstance.id_)\
                    .filter(ZoneInstance.zone_id == zone_sm.id_)\
                    .order_by(desc(ZoneInstance.mtime))\
                    .limit(zi_max_num)
            zi_query = db_session.query(ZoneInstance)\
                    .filter(ZoneInstance.zone_id == zone_sm.id_)\
                    .filter(ZoneInstance.id_ != zone_sm.zi_id)\
                    .filter(not_(ZoneInstance.id_.in_(zi_keep)))\
                    .filter(ZoneInstance.mtime < (func.now() - age_days))
            for zi in zi_query:
                if (zi.id_ == zone_sm.zi_id 
                        or zi.id_ == zone_sm.zi_candidate_id):
                    # Skip if this ZI has ben selected for republishing in 
                    # the mean time
                    continue
                db_session.delete(zi)
                count += 1
        result = {'num_deleted': count}
        self._finish_op()
        return result
   
    def vacuum_pare_deleted_zone_zis(self, age_days=None):
        """
        Pare ZIs on deleted zones older than age_days
        """
        self._begin_op()
        db_session = self.db_session
        db_query_slice = get_numeric_setting('db_query_slice', int)
        age_days_from_config = float(zone_cfg.get_row_exc(db_session, 
                                                    key='zone_del_pare_age'))
        if age_days_from_config <= 0 and age_days is None:
            return {'num_deleted': 0}
        if age_days is None:
            age_days = age_days_from_config
        age_days = timedelta(days=age_days)

        stmt = db_session.query(ZoneInstance.zone_id,
                func.count(ZoneInstance.id_).label('zi_count'))\
                        .group_by(ZoneInstance.zone_id).subquery()
        zone_sm_query = db_session.query(ZoneSM)\
                .filter(ZoneSM.state == ZSTATE_DELETED)\
                .outerjoin(stmt, ZoneSM.id_ == stmt.c.zone_id)\
                .filter(stmt.c.zi_count > 1)\
                .filter(and_(ZoneSM.deleted_start != None,
                    (func.now() - ZoneSM.deleted_start) > age_days))\
                .yield_per(db_query_slice)
        count = 0
        for zone_sm in zone_sm_query:
            zi_query = db_session.query(ZoneInstance)\
                    .filter(ZoneInstance.zone_id == zone_sm.id_)\
                    .filter(ZoneInstance.id_ != zone_sm.zi_id)
            if zone_sm.state != ZSTATE_DELETED:
                # Skip this if a customer has undeleted zone in the mean time..
                continue
            for zi in zi_query:
                if (zi.id_ == zone_sm.zi_id 
                        or zi.id_ == zone_sm.zi_candidate_id):
                    # Skip if this ZI has published or selected to be published
                    continue
                db_session.delete(zi)
                count += 1
        
        result = {'num_deleted': count}
        self._finish_op()
        return result

    def vacuum_syslog (self, age_days=None):
        """
        Destroy syslog messages received more than age_days ago
        """
        self._begin_op()
        db_session = self.db_session
        if age_days is None:
            age_days = float(zone_cfg.get_row_exc(db_session,
                key='syslog_max_age'))
        age_days = timedelta(days=age_days)
        count = 0
        # Do a straight SQL DELETE first to speed things along
        # Count events to be deleted
        syslog_table = sql_data['tables'][SyslogMsg]
        where_stmt = and_(SyslogMsg.receivedat != None,
                            (func.now() - SyslogMsg.receivedat) > age_days)
        count_select = select([func.count(syslog_table.c.get('id'))],
                                where_stmt)
        result = db_session.execute(count_select).fetchall()
        count += result[0][0]
        db_session.execute(syslog_table.delete().where(where_stmt))
                    
        result = {'num_deleted': count}
        self._finish_op()
        return result


    def _show_server(self, server_sm):
        """
        Show server backend
        """
        result = server_sm.to_engine(time_format=self.time_format)
        self._finish_op()
        return result

    def show_server(self, server_name):
        """
        Show a server, by name
        """
        self._begin_op()
        server_sm = find_server_byname(self.db_session, server_name)
        return self._show_server(server_sm)

    def show_server_byaddress(self, address):
        """
        Show a server, by address
        """
        self._begin_op()
        server_sm = find_server_byaddress(self.db_session, address)
        return self._show_server(server_sm)

    def create_server(self, server_name, address, sg_name=None,
                    server_type=None, ssh_address=None):
        """
        Create a Server SM
        """
        self._begin_op()
        new_server(self.db_session, server_name, address, sg_name, server_type,
                    ssh_address)
        self._finish_op()

    def delete_server(self, server_name):
        """
        Delete a Server SM
        """
        self._begin_op()
        del_server(self.db_session, server_name)
        self._finish_op()

    def set_server_ssh_address(self, server_name, ssh_address):
        """
        Perform set_server_ssh_address
        """
        self._begin_op()
        set_server_ssh_address(self.db_session, server_name, ssh_address)
        self._finish_op()
    
    def set_server(self, server_name, new_server_name=None,
                address=None, server_type=None, ssh_address=None):
        """
        Perform set_server
        """
        self._begin_op()
        set_server(self.db_session, server_name, new_server_name,
                address, server_type, ssh_address)
        self._finish_op()
    
    def rename_server(self, server_name, new_server_name=None,
                address=None, server_type=None):
        """
        Perform rename_server
        """
        self._begin_op()
        rename_server(self.db_session, server_name, new_server_name)
        self._finish_op()
    
    def move_server_sg(self, server_name, sg_name):
        """
        Move a server between SGs
        """
        self._begin_op()
        move_server_sg(self.db_session, server_name, sg_name)
        self._finish_op()

    def enable_server(self, server_name):
        """
        Enable a server
        """
        self._begin_op()
        db_session = self.db_session
        server_sm = find_server_byname(db_session, server_name)
        exec_server_sm(server_sm, ServerSMEnable) 
        self._finish_op()

    def disable_server(self, server_name):
        """
        Disable a server
        """
        self._begin_op()
        db_session = self.db_session
        server_sm = find_server_byname(db_session, server_name)
        exec_server_sm(server_sm, ServerSMDisable) 
        self._finish_op()

    def reset_server(self, server_name):
        """
        Reset server SM
        """
        self._begin_op()
        db_session = self.db_session
        server_sm = find_server_byname(db_session, server_name)
        exec_server_sm(server_sm, ServerSMReset) 
        self._finish_op()
    
    def write_rndc_conf(self):
        """
        Write out a new rndc.conf file
        """
        self._begin_op()
        db_session = self.db_session
        # Create temporary file for new rndc.conf
        rndc_conf_header = settings['rndc_header_template']
        rndc_conf_server = settings['rndc_server_template']
        rndc_conf_file = settings['rndc_conf_file']

        header_template = open(rndc_conf_header).readlines()
        header_template = ''.join(header_template)
        server_template = open(rndc_conf_server).readlines()
        server_template = ''.join(server_template)
        (fd, tmp_filename) = tempfile.mkstemp(
                    dir=settings['master_bind_config_dir'],
                    prefix='.' + basename(rndc_conf_file) + '-')
        tmp_file = io.open(fd, mode='wt')
        tmp_file.write(header_template)
        query = db_session.query(ServerGroup)
        for sg in query:
            for server_sm in sg.servers:
                # Also do disabled servers, as we want to be able to rndc them
                # when they are again enabled.
                filler = server_sm.to_engine()
                tmp_file.write(server_template % filler)
        tmp_file.close()
        # Rename tmp file into place so that replacement is atomic
        os.chown(tmp_filename, 0, 0)
        os.chmod(tmp_filename, int(settings['rndc_conf_file_mode'],8))
        os.rename(tmp_filename, rndc_conf_file)
        self._finish_op()

    def generate_tsig_key(self, file_name, key_name, 
            hmac_type='hmac-sha256'):
        """
        Generate a new tsig key in BIND named.conf format
        """
        if hmac_type.lower() not in tsig_key_algorithms:
            raise InvalidHmacType(algorithm)
        # Key size is optimally the full blocksize usablefor the hash in the
        # HMAC algorithm - RFC 2104
        if hmac_type.lower() in ('hmac-md5', 'hmac-sha1'):
            key_size = 64 #bytes
        elif hmac_type.lower() in ('hmac-sha224', 'hmac-sha256'):
            key_size = 128 # bytes
        elif hmac_type.lower() in ('hmac-sha384', 'hmac-512'):
            key_size = 256 #bytes
        else:
            key_size = 256 #bytes
        # Read key from /dev/random
        randev = open('/dev/random', 'rb')
        key_material = randev.read(key_size)
        randev.close()
        key_material = b64encode(key_material).decode()
        
        template_file = settings['tsig_key_template']
        key_template = open(template_file).readlines()
        key_template = ''.join(key_template)
        filler = {'key_name': key_name, 'algorithm': hmac_type, 
                'secret': key_material}
        key_text = key_template % filler

        old_umask = None
        write_file =  isinstance(file_name, str)
        if write_file:
            if file_name[0] != '/':
                file_name = (settings['dms_bind_config_dir'] 
                                + '/' + file_name)
            old_umask = os.umask(0o00077)
            output_file = open(file_name, 'wt') 
        else:
            output_file = sys.stdout
        print(key_text, file=output_file)
        output_file.flush()
        if write_file:
            os.chmod(file_name, int(settings['key_file_mode'],8))
            uid = pwd.getpwnam(settings['key_file_owner']).pw_uid
            gid = grp.getgrnam(settings['key_file_group']).gr_gid
            os.chown(file_name, uid, gid)
            output_file.close()
            os.umask(old_umask)
        return

    def rsync_server_admin_config(self, server_name, no_rndc=False):
        """
        Rsync configuration files to a server, and rndc reconfig it
        """
        self._begin_op()
        db_session = self.db_session
        server_sm = find_server_byname(db_session, server_name)
        config_dir = (settings['server_admin_config_dir'] + '/'
                            + server_sm.server_type)
        cmdline = (settings['rsync_path'] + ' ' + settings['rsync_args'] 
                + ' --password-file ' + settings['rsync_password_file'] 
                + ' ' + config_dir + '/' + ' ' + settings['rsync_target'])
        # Add IPv6 address squares
        address_string = '[' + server_sm.address + ']' \
                    if server_sm.address.find(':') else server_sm.address
        cmdline_str = cmdline % address_string
        cmdline = cmdline_str.split(' ')
        output = check_call(cmdline)
        if not no_rndc:
            cmdline = [settings['rndc_path'], '-s', server_sm.name, 'reconfig']
            output = check_call(cmdline)
        self._finish_op()

    def reset_all_zones(self):
        """
        Reset all zones
        """
        self._begin_op()
        db_session = self.db_session
        id_query = db_session.query(ZoneSM.id_, ZoneSM.name)
        id_query = ZoneSM.query_is_not_disabled_deleted(id_query)
        id_result = id_query.all()
        for zone_id, zone_name in id_result:
            try:
                zone_sm = db_session.query(ZoneSM)\
                        .filter(ZoneSM.id_ == zone_id).one()
            except NoResultFound:
                raise ZoneNotFoundByZoneId(zone_id)
            exec_zonesm(zone_sm, ZoneSMDoReset)
        self._finish_op()

    def list_resolv_zi_id(self, name, zi_id):
        """
        Extra functionality for zone_tool ls_zi command.  Allows ls_zi to take
        a zi_id argument
        """
        self._begin_op()
        zone_sm = self._get_zone_sm(name)
        zi = self._resolv_zi_id(zone_sm, zi_id)
        if not zi:
            raise ZiNotFound(name, zi_id)
        resolv_result = zi.to_engine_brief(time_format=self.time_format)
        result = {'all_zis': [resolv_result], 'zi_id': zone_sm.zi_id}
        self._finish_op()
        return result

    def restore_named_db(self):
        """
        Dump dms DB to Named zone files and include file

        For quick DR secnario
        """
        self._begin_op()
        
        # Exception processing make the both of the following a bit of a 
        # rats nest
        # Check that named is not running
        cmdline = [settings['rndc_path'], 'status']
        try:
            check_output(cmdline, stderr=STDOUT)
        except CalledProcessError as exc:
            if exc.returncode != 1:
                raise exc
            pass
        else:
            raise NamedStillRunning(0)

        # Check that dmsdmd is not running
        try:
            pid_file = open(settings['pid_file'],'r')
            dmsdmd_pid = int(pid_file.readline().strip())
            pid_file.close()
            # The following throws exceptions if process does not exist etc!
            # Sending signal 0 does not touch process, but  call succeeds
            # if it exists
            os.kill(dmsdmd_pid, 0)
        except ValueError as exc:
            # Error from int() type conversion above
            raise PidFileValueError(pid_file, exc)
        except (IOError,OSError) as exc:
            if (exc.errno in (errno.ESRCH,)):
                # This is from kill()
                raise DmsdmdStillRunning(dmsdmd_pid)
            # File IO causes this
            elif (exc.errno in (errno.ENOENT,)):
                # This file may be removed by dameon nicely shutting down.
                pass
        else:
            # No exceptions, dmsdmd is running!!!
            raise DmsdmdStillRunning(dmsdmd_pid)

        # Dump out each zone file
        db_session = self.db_session
        id_query = db_session.query(ZoneSM.id_, ZoneSM.name)
        id_query = ZoneSM.query_is_not_disabled_deleted(id_query)
        id_result = id_query.all()
        for zone_id, zone_name in id_result:
            try:
                zone_sm = db_session.query(ZoneSM)\
                        .filter(ZoneSM.id_ == zone_id).one()
            except NoResultFound:
                raise ZoneNotFoundByZoneId(zone_id)
            try:
                zone_sm.write_zone_file(db_session, ZoneFileWriteInternalError)
                zone_sm.state = ZSTATE_PUBLISHED
                db_session.commit()
            except ZoneFileWriteInternalError as exc:
                db_session.rollback()
                raise ZoneFileWriteError(str(exc))
        
        # Write out config file include
        master_sm = get_master_sm(db_session)
        try:
            master_sm.write_named_conf_includes(db_session, 
                    NamedConfWriteInternalError)
        except NamedConfWriteInternalError as exc:
            raise NamedConfWriteError(str(exc))
        self._finish_op()

    def list_failed_events(self, last_limit=None):
        """
        List failed events
        """
        self._begin_op()
        if not last_limit:
            last_limit = get_numeric_setting('list_events_last_limit', 
                                                float)
        db_query_slice = get_numeric_setting('db_query_slice', int)
        db_session = self.db_session
        query = db_session.query(Event).filter(Event.state == ESTATE_FAILURE)\
                .order_by(desc(Event.id_)).limit(last_limit)\
                .yield_per(db_query_slice)
        results = []
        for event in query:
            json_event = event.to_engine_brief(time_format=self.time_format)
            results.append(json_event)
        self._finish_op()
        return results

    def list_events(self, last_limit=None):
        """
        List failed events
        """
        self._begin_op()
        if not last_limit:
            last_limit = get_numeric_setting('list_events_last_limit', 
                                                float)
        db_query_slice = get_numeric_setting('db_query_slice', int)
        db_session = self.db_session
        query = db_session.query(Event)\
                .order_by(desc(Event.id_)).limit(last_limit)\
                .yield_per(db_query_slice)
        results = []
        for event in query:
            json_event = event.to_engine_brief(time_format=self.time_format)
            results.append(json_event)
        self._finish_op()
        return results

    def _find_event(self, event_id):
        """
        Find an event by ID
        """
        db_session = self.db_session
        try:
            query = db_session.query(Event).filter(Event.id_ == event_id)
            event = query.one()
        except NoResultFound as exc:
            raise EventNotFoundById(event_id)
        return event

    def show_event(self, event_id):
        """
        Show an event
        """
        self._begin_op()
        event = self._find_event(event_id)
        result = event.to_engine(time_format = self.time_format)
        self._finish_op()
        return result

    def fail_event(self, event_id):
        """
        Fail an event
        """
        self._begin_op()
        event = self._find_event(event_id)
        if not event.state in (ESTATE_NEW, ESTATE_RETRY):
            raise CantFailEventById(event_id)
        cancel_event(event.id_, self.db_session)
        self._finish_op()




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Resource Record comment class, corresponding to rr_comments table
"""

from magcode.core.database import *
from dms.dns import RRCLASS_IN

@saregister
class RRComment(object):
    """
    DNS Resource Record comment.
    """
    _table="rr_comments"

    def __init__(self, comment=None, tag=None):
        """
        Initialise a resource record comment

        Assign any id_ after creating instance if needed as 
        SQLAlchemy usually sets it up from DB
        """
        self.comment = comment
        self.tag = tag

    def to_engine_brief(self, time_format=None):
        """
        Output for zone engine.
        """
        return{'comment_id': self.id_, 'comment': self.comment,
                'tag': self.tag}

    to_engine = to_engine_brief





#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
DMS DB Zone Query Support
"""


from sqlalchemy.orm.exc import NoResultFound 

from magcode.core.database import *


def rr_query_db_raw(db_session, label, name=None, type_=None, zi_id=None,
        rdata=None, include_disabled=False, sectag=None):
    """
    Function to query Zone DB, and return records
    """
    # We get types from sql_types to avoid import nesting problems
    ZoneSM = sql_types['ZoneSM']
    ResourceRecord = sql_types['ResourceRecord']
    query_kwargs = {'label': label, 'name': name, 'type': type_, 'rdata': rdata,
            'zi_id': zi_id, 'include_disabled': include_disabled}
    domain = name
    zone_sm = None
    # Adjust domain and label if needed
    if not domain:
        # Walk up label from root '.' to obtain most specific Zone in DB
        # Stip dot to label if it already does not have one
        if  not label.endswith('.'):
            label = label + '.'
        label = label.lower()
        node_labels = label.split('.')
        node_labels.reverse()
        node_labels = node_labels[1:]
        test_domain = ''
        for l in node_labels:
            # Find most specific domain in DB by accumlating the node labels
            test_domain = l + '.' + test_domain
            if test_domain == '.':
                # Skip root domain
                continue
            try:
                query = db_session.query(ZoneSM)\
                        .filter(ZoneSM.name == test_domain)
                if not include_disabled:
                    query = ZoneSM.query_is_not_disabled_deleted(query)
                else:
                    query = ZoneSM.query_is_not_deleted(query)
                zone_sm = query.one()
            except NoResultFound:
                continue
            if(sectag and zone_sm and sectag.sectag != settings['admin_sectag'] 
                    and sectag not in zone_sm.sectags):
                continue
            domain = test_domain
       
        if not zone_sm:
            return None
        d_index = label.rfind(domain)
        if d_index == 1:
            raise ValueError("Domain must not start with a '.'")
        elif (d_index > 1):   
            label = label[:(d_index-1)]
        else:
            label = '@'
    else:
        # Check input
        if not domain.endswith('.'):
            domain += '.'
        if domain[0] == '.':
            raise ValueError("Domain must not start with a '.'")
        domain = domain.lower()
        label = label.lower()
        if not label:
            label = '@'
        # Check if domain exists
        try:
            query = db_session.query(ZoneSM)\
                        .filter(ZoneSM.name == domain)
            if not include_disabled:
                query = ZoneSM.query_is_not_disabled_deleted(query)
            else:
                query = ZoneSM.query_is_not_deleted(query)
            zone_sm = query.one()
        except NoResultFound:
            return None
        if(sectag and zone_sm and sectag.sectag != settings['admin_sectag'] 
                and sectag not in zone_sm.sectags):
            return None
    # Now we have a valid domain, time to match the records within it 
    # Replace wildcards in label with SQL wild cards
    label = label.replace('*', '%')
    label = label.replace('?', '_')
    # Perform Query
    if zi_id == None:
        zi_id = zone_sm.zi.id_
    query = db_session.query(ResourceRecord)\
            .filter(ResourceRecord.label.like(label))
    if zi_id:
        if isinstance(zi_id, tuple) or isinstance(zi_id, list):
            query = query.filter(ResourceRecord.zi_id.in_(zi_id))
        else:
            query = query.filter(ResourceRecord.zi_id == zi_id)
    if not include_disabled:
        query = query.filter(ResourceRecord.disable == False)
    if type_:
        if isinstance(type_, tuple) or isinstance(type_, list):
            type_ = [ t.upper() for t in type_ ]
            query = query.filter(ResourceRecord.type_.in_(type_))
        else:
            query = query.filter(ResourceRecord.type_ == type_.upper())
    if rdata:
        query = query.filter(ResourceRecord.rdata == rdata)
    rrs = query.all()
    if not len(rrs):
        return None
    label = label.replace('%', '*')
    label = label.replace('_', '?')
    result = {'query': query_kwargs, 'label': label, 'name': zone_sm.name, 
            'zone_sm': zone_sm, 'zi_id': zi_id, 'rrs': rrs}
    return result





#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Configuration State Machine

There is some verbose programming in here due to similar sections, but it is
better to lay everything out so that you can see whats happening.
"""

from tempfile import mkstemp
import io
import os
import grp
import pwd
import socket
from os.path import basename
from datetime import timedelta
from random import random
from subprocess import check_call
from subprocess import CalledProcessError

from sqlalchemy.sql import or_
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.orm import relationship

from magcode.core.utility import get_numeric_setting
from magcode.core.utility import get_configured_addresses
from magcode.core.database import *
from magcode.core.database.state_machine import StateMachine
from magcode.core.database.state_machine import smregister
from magcode.core.database.state_machine import SMEvent
from magcode.core.database.state_machine import SMSyncEvent
from magcode.core.database.state_machine import StateMachineError
from magcode.core.database.state_machine import StateMachineFatalError
from magcode.core.database.event import eventregister
from magcode.core.database.event import synceventregister
from magcode.core.database.event import create_event
from magcode.core.database.event import ESTATE_SUCCESS
from dms.globals_ import MASTER_SERVER_ACL_TEMPLATE
from dms.template_cache import read_template
from dms.template_cache import clear_template_cache
from dms.exceptions import ReplicaSgExists


# Some constants
MSTATE_HOLD = "HOLD"
MSTATE_READY = "READY"

HOLD_SG_NONE = 0
HOLD_SG_MASTER = -1
HOLD_SG_ALL = -2

# Configuration State Machine Exceptions
class MasterConfigFileError(StateMachineFatalError):
    """
    Can not write configuration file, or cannot access a master template file,
    or there was a template key error.
    """
class ServerConfigFileError(StateMachineFatalError):
    """
    Can not write configuration file, or cannot access a server template file,
    or there was a template key error.
    """
class CantContactMasterServer(StateMachineError):
    """
    rndc operation, can't contact master dns server
    """
class CantFindSg(StateMachineFatalError):
    """
    rndc operation, can't contact master dns server
    """
class RndcFatalError(StateMachineFatalError):
    """
    rndc operation, unrecognised exit code.
    """
@eventregister
class MasterSMLoadKeys(SMEvent):
    """
    rndc operation, Master Server only, DNSSEC loadkeys
    """
# Commenting out as duplicate declaration casuing SAWarnings.  One below
# is the one being actually used in state machines....
#@eventregister
#class MasterSMSignZone(SMSyncEvent):
#    """
#    rndc operation, Master Server only, DNSSEC sign zone
#
#    Can either be synchronous or scheduled.
#    """
@eventregister
class MasterSMSignZone(SMEvent):
    """
    rndc operation, Master Server only, DNSSEC sign zone
    """
@eventregister
class MasterSMMasterReconfig(SMEvent):
    """
    Configuration update, Master Server only.

    Useful for DNSSEC zone transitions
    """
@eventregister
class MasterSMPartialReconfig(SMEvent):
    """
    Configuration update, Master and one Server Group
    """
@eventregister
class MasterSMAllReconfig(SMEvent):
    """
    Configuration update, Master and All Server Groups
    """
@eventregister
class MasterSMHoldTimeout(SMEvent):
    """
    Hold  time out event
    """
@eventregister
class MasterSMReset(SMEvent):
    """
    Reset Config SM and queue a MasterSMAllReconfig
    """
@synceventregister
class MasterSMBatchHold(SMSyncEvent):
    """
    Hold  time out event
    """

@smregister
class MasterSM(StateMachine):
    """
    Configuration State Machine
    """
    _table = 'sm_master'
    _sm_events = (MasterSMLoadKeys, MasterSMSignZone, MasterSMMasterReconfig,
            MasterSMPartialReconfig, MasterSMAllReconfig, MasterSMHoldTimeout,
            MasterSMBatchHold,MasterSMReset)
    
    @classmethod
    def _mapper_properties(class_):
        ServerGroup = sql_types['ServerGroup']
        ServerSM = sql_types['ServerSM']
        return {'replica_sg': relationship(ServerGroup, backref='master_sm'),
                'master_server': relationship(ServerSM, 
                                        backref='master_sm')}

    def _init(self):
        self.hold_sg = HOLD_SG_NONE
        self.hold_sg_name = None
        self.hold_start = None
        self.hold_stop = None
        self.state = MSTATE_READY
    
    def __init__(self):
        self._init()
        # These should only be initialised on initial MasterSM creation
        self.replica_sg_id = None
        self.master_server_id = None
        self.master_hostname = None

    def write_named_conf_includes(self, db_session, op_exc):
        """
        Write the bits of named configuration.

        Seperated so that it is callable from recovery script
        """
        def open_tmp_file(prefix):
            (fd, tmp_filename) = mkstemp(
                        dir=tmp_dir,
                        prefix=prefix)
            include_file = io.open(fd, mode='wt')
            return (include_file, tmp_filename)

        def clean_up_rename(include_file, tmp_filename, config_file_name):
                include_file.close()
                # Rename tmp file into place so that replacement is atomic
                run_as_user = settings['run_as_user']
                try:
                    run_as_user_pwd = pwd.getpwnam(run_as_user)
                except KeyError as exc:
                    msg = ("Could not find user '%s' in passwd database - %s"
                                % (run_as_user, str(exc)))
                    raise op_exc(msg)
                uid = run_as_user_pwd.pw_uid 
                gid = run_as_user_pwd.pw_gid 
                os.chown(tmp_filename, uid, gid)
                os.chmod(tmp_filename, int(settings['zone_file_mode'],8))
                # Rename tmp file into place so that replacement is atomic
                os.chmod(tmp_filename, int(settings['config_file_mode'],8))
                os.rename(tmp_filename, config_file_name)

        db_query_slice = get_numeric_setting('db_query_slice', int)
        # Clear template cache.  This forces a re read of all templates
        clear_template_cache()
        # Rewrite include and global server ACL file if required.
        # Trap file IO errors as event queue can't handle them.
        try:
            tmp_dir = settings['master_config_dir']
            # master server ACL file
            acl_prefix = ('.' 
                    + basename(settings['master_server_acl_file']) + '-')
            acl_file, tmp_filename = open_tmp_file(acl_prefix)

            # Create Master ACL file 
            server_acl_template = read_template(
                            settings['master_template_dir'] + '/' 
                            + settings[MASTER_SERVER_ACL_TEMPLATE])
            server_acls = {}
            ServerGroup = sql_types['ServerGroup']
            query = db_session.query(ServerGroup)
            for sg in query:
                # Each SG gets its own ACL to prevent cross  SG
                # domain discovery if a server is compromised.
                sg_acl_name = sg.name + settings['acl_name_extension']
                server_acls[sg.name] = {'acl_name': sg_acl_name, 
                                        'servers': ''}
                for server_sm in sg.servers:
                    # include disabled server, as access can be shut off
                    # in IPSEC and firewall!
                    server_acls[sg.name]['servers'] += ("%s;\n" 
                                                    % server_sm.address)
                    del server_sm
                if not server_acls[sg.name]['servers']:
                    server_acls[sg.name]['servers'] = 'none;\n'
                # Stop memory leaks
                del sg
            for sg_name in server_acls:
                acl_file.write(server_acl_template % server_acls[sg_name])
            clean_up_rename(acl_file, tmp_filename, 
                    settings['master_server_acl_file'])

            # include file
            include_prefix = ('.' 
                    + basename(settings['master_include_file']) + '-')
            include_file, tmp_filename = open_tmp_file(include_prefix)
            # Get list of zones from zone_sm, and write out each 
            # config file section
            ZoneSM = sql_types['ZoneSM']
            query = ZoneSM.query_is_configured(
                        db_session.query(ZoneSM)).yield_per(db_query_slice)
            for zone_sm in query:
                zone_sm.write_config(include_file, db_session, server_acls, 
                        self.replica_sg)
                del zone_sm
            clean_up_rename(include_file, tmp_filename, 
                    settings['master_include_file'])
        except (IOError, OSError) as exc: 
            msg = ( "Could not access/write file '%s' - %s." 
                            % (exc.filename, exc.strerror))
            raise op_exc(msg)
        except KeyError as exc:
            msg = ("Invalid template key in template dir %s - %s"
                    % (settings['master_template_dir'], str(exc)))
            raise op_exc(msg)
        finally:
            # clean up if possible
            try:
                os.unlink(tmp_filename)
            except:
                pass

    def _master_rndc(self, event, *rndc_args):
        """
        Write out include file for Master server

        This is done as part of the Master Server Config SM unlike the server 
        servers, where it is part of the SG code, and is exceuted by the 
        MasterSM.
        """
        if (rndc_args[0] == 'reconfig' 
                or (rndc_args[0] == 'reload' and len(rndc_args) == 1)):
            db_session = event.db_session
            self.write_named_conf_includes(db_session, MasterConfigFileError)

        # Run rndc 
        try:
            cmdline = [settings['rndc_path']]
            cmdline.extend(rndc_args)
            output = check_call(cmdline)
        except CalledProcessError as exc:
            if exc.returncode == 1:
                msg = (
                    "%s could not contact master DNS server, return code '%s'" 
                   % (settings['rndc_path'], exc.returncode))
                raise CantContactMasterServer(msg)
            else:
                msg = str(exc)
                raise RndcFatalError(msg)

        return (RCODE_OK, "'%s' on master DNS server completed" 
                                    % ' '.join(cmdline)) 

    def _rndc_dnssec(self, event, operation):
        """
        rndc loadkeys/sign zone only
        """
        db_session = event.db_session
        sm_id = event.py_parameters['sm_id']
        zone_name = event.py_parameters['zone_name']
        rndc_args = [operation, zone_name]
        return self._master_rndc(event, *rndc_args)

    def _rndc_load_keys(self, event):
        """
        rndc loadkeys zone
        """
        return self._rndc_dnssec(event, 'loadkeys')

    def _rndc_sign_zone(self, event):
        """
        rndc loadkeys zone
        """
        return self._rndc_dnssec(event, 'sign')

    def _hold_enter(self, db_session, all_reconfig=False):
        """
        Enter hold state

        Set fields as needed.
        """
        if self.state == MSTATE_HOLD:
            return
        self.hold_sg = HOLD_SG_ALL if all_reconfig else HOLD_SG_NONE
        self.state = MSTATE_HOLD
        self.hold_start = db_time(db_session)
        delay=timedelta(minutes=float(settings['master_hold_timeout']))
        self.hold_stop = self.hold_start + delay
        # Queue hold timeout
        create_event(MasterSMHoldTimeout, db_session=db_session,
                sm_id=self.id_, master_id=self.id_, delay=delay)

    def _reset(self, event):
        """
        Reset mastersm
        """
        self._init()
        # Queue all reconfig
        create_event(MasterSMAllReconfig, db_session=event.db_session,
                sm_id=self.id_, master_id=self.id_)
        return (RCODE_OK, "MasterSM - SM reinitialised and MasterSMAllReconfig")

    def _batch_hold(self, event):
        """
        Process a batch hold start event

        Made to be used from both states
        """
        self._hold_enter(event.db_session, all_reconfig=True)
        return (RCODE_OK, "MasterSM - CONFIG_HOLD via MasterSMBatchHold event")

    def _ready_master_reconfig(self, event):
        """
        Process a master only reconfig
        """
        # Update server address info
        recalc_machine_dns_server_info(event.db_session)
        # Update master server configuration
        rcode, msg = self._master_rndc(event, 'reconfig')
        if rcode != RCODE_OK:
            return (rcode, msg)
        self._hold_enter(event.db_session)
        return (RCODE_OK, "MasterSM: master reconfig only done")

    def _ready_partial_reconfig(self, event):
        """
        Process a partial reconfig event
        """
        # Update server address info
        recalc_machine_dns_server_info(event.db_session)
        # Update master server configuration
        rcode, msg = self._master_rndc(event, 'reconfig')
        if rcode != RCODE_OK:
            return (rcode, msg)

        # Issue reconfig event to SG
        db_session = event.db_session
        ServerGroup = sql_types['ServerGroup']
        # sg_id found in zone_sm and sent here as event parameter
        sg_id = event.py_parameters['sg_id']
        sg_name = event.py_parameters['sg_name']
        try:
            sg = db_session.query(ServerGroup)\
                    .filter(ServerGroup.id_ == sg_id).one()
        except NoResultFound as exc:
            msg = ("MasterSM: can't find SG %s by id '%s'" 
                % (sg_name, sg_id))
            raise CantFindSg(msg)
        
        self.hold_sg_name = sg.name
        
        delay_time = timedelta(
                seconds=float(settings['master_rndc_settle_delay']))
        # Replica SG reconfigure
        replica_sg = self.replica_sg
        if (replica_sg and replica_sg is not sg):
            try:
                replica_sg.write_config(db_session, ServerConfigFileError)
            except ServerConfigFileError as exc:
                log_error(str(exc))
            else:
                for server_sm in replica_sg.servers:
                    if server_sm.is_disabled():
                        continue
                    create_event(sql_events['ServerSMConfigChange'],
                                db_session=event.db_session,
                                sm_id=server_sm.id_, server_id=server_sm.id_,
                                delay=delay_time,
                                server_name=server_sm.name)

        # SG reconfigure
        try:
            sg.write_config(db_session, ServerConfigFileError)
        except ServerConfigFileError as exc:
            log_error(str(exc))
        else:
            # Reconfigure servers in this group
            for server_sm in sg.servers:
                if server_sm.is_disabled():
                    continue
                create_event(sql_events['ServerSMConfigChange'],
                            db_session=event.db_session,
                            sm_id=server_sm.id_, server_id=server_sm.id_,
                            delay=delay_time,
                            server_name=server_sm.name)

        self._hold_enter(db_session)
        return (RCODE_OK, "MasterSM: SG '%s' - master named reconfig done and SG reconfig queued"
                            % self.hold_sg_name)

    def _ready_all_reconfig(self, event):
        """
        Process an all reconfig
        """
        # Update server address info
        recalc_machine_dns_server_info(event.db_session)
        # Update master server configuration
        rcode, msg = self._master_rndc(event, 'reconfig')
        if rcode != RCODE_OK:
            return (rcode, msg)
        
        # Issue all reconfig event to all servers
        delay_time = timedelta(
                seconds=float(settings['master_rndc_settle_delay']))
        db_session = event.db_session
        ServerGroup = sql_types['ServerGroup']
        for sg in db_session.query(ServerGroup):
            try:
                sg.write_config(db_session, ServerConfigFileError)
            except ServerConfigFileError as exc:
                log_error(str(exc))
                continue
            for server_sm in sg.servers:
                if server_sm.is_disabled():
                    continue
                create_event(sql_events['ServerSMConfigChange'],
                            db_session=event.db_session,
                            sm_id=server_sm.id_, server_id=server_sm.id_,
                            delay=delay_time,
                            server_name=server_sm.name)
        
        self._hold_enter(db_session)
        return (RCODE_OK, 
                "Master named reconfig done and reconfig queued for all SGs")
        
    def _hold_master_reconfig(self, event):
        """
        Process a master reconfig event during hold

        Sets master_sm hold level as apropriate
        """
        db_session = event.db_session
        sm_id = event.py_parameters['sm_id']

        if self.hold_sg == HOLD_SG_NONE:
            self.hold_sg = HOLD_SG_MASTER
            self.hold_sg_name = None
        return (RCODE_OK, 
                "MasterSM - hold event %s, hold_sg now '%s'"
                    % (MasterSMMasterReconfig.__name__, 
                        self._display_hold_sg()))

    def _hold_partial_reconfig(self, event):
        """
        Process a reconfig event during hold

        Sets master_sm hold level as apropriate
        """
        db_session = event.db_session
        sm_id = event.py_parameters['sm_id']
        sg_id = event.py_parameters['sg_id']
        sg_name = event.py_parameters['sg_name']

        if self.hold_sg in (HOLD_SG_NONE, HOLD_SG_MASTER):
            self.hold_sg = sg_id
            self.hold_sg_name = sg_name
        elif self.hold_sg == HOLD_SG_ALL:
            pass
        elif self.hold_sg != sg_id:
            self.hold_sg = HOLD_SG_ALL
            self.hold_sg_name = None

        return (RCODE_OK, 
            "MasterSM - hold event %s, hold_sg now '%s',"
                " hold_sg_name '%s'"
                    % (MasterSMPartialReconfig.__name__, 
                        self._display_hold_sg(), self.hold_sg_name))

    def _hold_all_reconfig(self, event):
        """
        Process an all reconfig event during hold

        Sets master_sm hold_sg to HOLD_SG_ALL
        """
        db_session = event.db_session
        sm_id = event.py_parameters['sm_id']

        # OK, its eveything!
        self.hold_sg = HOLD_SG_ALL
        self.hold_sg_name = None

        return (RCODE_OK, 
                "MasterSM -  hold event %s, hold_sg now '%s'"
                    % (MasterSMAllReconfig.__name__, self._display_hold_sg()))

    def _hold_time_out(self, event):
        """
        Process a hold time out.

        This the event runs the associated SM backend routine depending on 
        value of self.hold_sg
       """
        db_session = event.db_session
        sm_id = event.py_parameters['sm_id']
        old_hold_sg = self.hold_sg
        old_hold_sg_name = self.hold_sg_name
        # Reset All SM fields
        self._init()
        if old_hold_sg == HOLD_SG_ALL:
            create_event(MasterSMAllReconfig, 
                    sm_id = self.id_, master_id = self.id_)
            return(RCODE_OK, 
                    "MasterSM - %s, %s created and queued"
                    % (MasterSMHoldTimeout.__name__, 
                        MasterSMAllReconfig.__name__))
        if old_hold_sg == HOLD_SG_MASTER:
            create_event(MasterSMMasterReconfig, 
                    sm_id = self.id_, master_id = self.id_)
            return(RCODE_OK,
                "MasterSM - %s, %s created and queued"
                % (MasterSMHoldTimeout.__name__, 
                    MasterSMMasterReconfig.__name__))
        elif old_hold_sg:
            create_event(MasterSMPartialReconfig, 
                    sm_id = self.id_, master_id = self.id_,
                    sg_id = old_hold_sg, sg_name = old_hold_sg_name)
            return(RCODE_OK,
        "MasterSM - %s, %s for SG %s(%s) created and queued"
            % (MasterSMHoldTimeout.__name__, 
                MasterSMPartialReconfig.__name__, 
                old_hold_sg_name, self._display_hold_sg(old_hold_sg)))
                
        return(RCODE_NOCHANGE, "MasterSM - no reconfigure event during hold")

    _sm_table = { MSTATE_READY: { 
                        MasterSMMasterReconfig: _ready_master_reconfig,
                        MasterSMPartialReconfig: _ready_partial_reconfig,
                        MasterSMAllReconfig: _ready_all_reconfig,
                        MasterSMBatchHold: _batch_hold,
                        MasterSMLoadKeys: _rndc_load_keys,
                        MasterSMSignZone: _rndc_sign_zone,
                        MasterSMReset: _reset,
                        },
                  MSTATE_HOLD: {
                        MasterSMHoldTimeout: _hold_time_out,
                        MasterSMMasterReconfig: _hold_master_reconfig,
                        MasterSMPartialReconfig: _hold_partial_reconfig,
                        MasterSMAllReconfig: _hold_all_reconfig,
                        MasterSMBatchHold: _batch_hold,
                        MasterSMLoadKeys: _rndc_load_keys,
                        MasterSMSignZone: _rndc_sign_zone,
                        MasterSMReset: _reset,
                        },
                  }

    def _display_hold_sg(self, hold_sg=None):
        """
        Get display value for hold_sg
        """
        if not hold_sg:
            hold_sg = self.hold_sg
        if hold_sg == HOLD_SG_ALL:
            display = 'HOLD_SG_ALL'
        elif hold_sg == HOLD_SG_NONE:
            display = 'HOLD_SG_NONE'
        elif hold_sg == HOLD_SG_MASTER:
            display = 'HOLD_SG_MASTER'
        else:
            display = '%s' % hold_sg
        return display

    def to_engine_brief(self, time_format=None):
        """
        Brief dict of master_sm fields
        """
        return {'master_id': self.id_, 'state': self.state}

    def to_engine(self, time_format=None):
        """
        Dict of master_sm fields
        """
        hold_sg = self._display_hold_sg()
        if not time_format:
            hold_start = (self.hold_start.isoformat(sep=' ')
                            if self.hold_start else None)
            hold_stop = (self.hold_stop.isoformat(sep=' ')
                            if self.hold_stop else None)
        else:
            hold_start = (self.hold_start.strftime(time_format)
                            if self.hold_start else None)
            hold_stop = (self.hold_stop.strftime(time_format)
                            if self.hold_stop else None)
        return {'master_id': self.id_, 'state': self.state, 
                'hold_start': hold_start, 'hold_stop':hold_stop, 
                'hold_sg': hold_sg, 'hold_sg_name': self.hold_sg_name,
                'master_server_id': self.master_server_id,
                'master_server': self.master_server.name if self.master_server
                                    else self.master_hostname,
                'replica_sg_id': self.replica_sg_id, 
                'replica_sg_name': self.replica_sg.name if self.replica_sg 
                                    else None }

def get_master_sm(db_session):
    """
    get master_sm from database, create it if not there
    """
    try:
        master_sm = db_session.query(MasterSM).one()
    except MultipleResultsFound as exc:
        # Blow up REAL BIG!
        log_critical("More than one MasterSM found in database, giving up")
        systemd_exit(os.EX_SOFTWARE, SDEX_GENERIC)
    except NoResultFound:
        master_sm = MasterSM()
        db_session.add(master_sm)
        db_session.flush()

    return master_sm

def batch_hold(db_session):
    """
    Start a batch hold zone creation state
    """
    master_sm = get_master_sm(db_session)
    batch_hold_event = MasterSMBatchHold()
    results = batch_hold_event.execute()
    if results['state'] != ESTATE_SUCCESS:
        raise ConfigBatchHoldFailed()

def zone_sm_reconfig_schedule(db_session, zone_sm, zone_sm_event=None,
                            randomize=False, master_reconfig=False, **kwargs):
    """
    Schedule MasterSM zone creation/update events

    Zone SM helper function
    """
    master_sm = get_master_sm(db_session)
    # According to sampling theorem, twice tick rate, add 1 to account for
    # safety
    coalesce_time = timedelta(seconds=3*float(settings['sleep_time']))
    # Make delay_secs 2 * coalesce time
    delay_secs = 6*float(settings['sleep_time'])
    if randomize:
        delay_secs += 60*float(settings['master_hold_timeout']) * random()
    delay_time = timedelta(seconds=delay_secs)
    sg_id = zone_sm.sg.id_
    # Queue config event
    if master_reconfig:
        if (master_sm.state != MSTATE_HOLD
            or (master_sm.state == MSTATE_HOLD
                and master_sm.hold_sg == HOLD_SG_NONE)):
            create_event(MasterSMMasterReconfig, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_)
    # Only create MasterSMPartialEvents when needed, as they clog the 
    # event queue
    elif (master_sm.state != MSTATE_HOLD
            or (master_sm.state == MSTATE_HOLD 
                and (master_sm.hold_sg != sg_id 
                    and master_sm.hold_sg != HOLD_SG_ALL))):
        create_event(MasterSMPartialReconfig, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_,
                        sg_id=sg_id, sg_name=zone_sm.sg.name)
    # Queue zone_sm event
    if not zone_sm_event:
        return
    if master_sm.state == MSTATE_READY:
        create_event(zone_sm_event, db_session=db_session,
                        sm_id=zone_sm.id_, zone_id=zone_sm.id_,
                        name=zone_sm.name,
                        delay=delay_time, coalesce_period=coalesce_time,
                        **kwargs)
        return
    elif master_sm.state == MSTATE_HOLD:
        schedule_time = master_sm.hold_stop + delay_time
        create_event(zone_sm_event, db_session=db_session,
                        time=schedule_time, coalesce_period=coalesce_time,
                        sm_id=zone_sm.id_, zone_id=zone_sm.id_,
                        name=zone_sm.name, **kwargs)
        return
    else:
        log_critical('MasterSM - unrecognized state, exiting')
        systemd_exit(os.EX_SOFTWARE, SDEX_GENERIC)
    return

def zone_sm_dnssec_schedule(db_session, zone_sm, operation):
    """
    Schedule a DNSSEC rndc sign/loadkeys operation for a zone_sm
    """
    master_sm = get_master_sm(db_session)
    if operation == 'loadkeys':
        create_event(MasterSMLoadKeys, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_,
                        zone_name=zone_sm.name)
    elif operation in ('sign', 'signzone'):
        create_event(MasterSMSignZone, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_,
                        zone_name=zone_sm.name)
    else:
        log_error("MasterSM - zone '%s', invalid dnssec operation"
                % zone_sm.name)

def show_master_sm(db_session, time_format=None):
    """
    Return a dict consisting of the MasterSM
    """
    master_sm = get_master_sm(db_session)
    return master_sm.to_engine(time_format)

def reset_master_sm(db_session):
    """
    Reset the Configuration state machine
    """
    master_sm = get_master_sm(db_session)
    create_event(MasterSMReset, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_)

def reconfig_all(db_session):
    """
    Reconfigure all DNS servers - helper
    """
    master_sm = get_master_sm(db_session)
    create_event(MasterSMAllReconfig, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_)

def reconfig_sg(db_session, sg_id, sg_name):
    """
    Reconfigure An SGs DNS servers - helper
    """
    master_sm = get_master_sm(db_session)
    create_event(MasterSMPartialReconfig, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_,
                        sg_id=sg_id, sg_name=sg_name)

def reconfig_master(db_session):
    """
    Reconfigure Master DNS server
    """
    master_sm = get_master_sm(db_session)
    create_event(MasterSMMasterReconfig, db_session=db_session,
                        sm_id=master_sm.id_, master_id=master_sm.id_)

def set_mastersm_replica_sg(db_session, sg):
    """
    Set the replica SG
    """
    master_sm = get_master_sm(db_session)
    if hasattr(master_sm, 'replica_sg') and master_sm.replica_sg and sg:
        raise ReplicaSgExists(sg.name, master_sm.replica_sg.name)
    master_sm.replica_sg = sg
    # This is being done straight after this call....
    # db_session.flush()

def get_mastersm_replica_sg(db_session):
    """
    Get the replica SG
    """
    master_sm = get_master_sm(db_session)
    return master_sm.replica_sg

def get_mastersm_master_server(db_session):
    """
    Get the master server setting, if it exists
    """
    if hasattr(self, 'master_server') and self.master_server:
        return master_server
    return None

def recalc_machine_dns_server_info(db_session, ifconfig_exc=False):
    """
    Recalculate DNS server connection information for this machine
    """
    # Get machines configurede addresses via 'ip addr' (Linux) 
    # or 'ifconfig -a' (FreeBSD, *BSD?)
    try:
        configured_addresses = get_configured_addresses()
    except CalledProcessError as exc:
        if ifconfig_exc:
            raise(exc)
        log_error(str(exc))
        return
    # Traverse server_groups table and add all connectable master_address and 
    # master_alt_address to this_servers_addresses
    this_servers_addresses = []
    ServerGroup = sql_types['ServerGroup']
    for sg in db_session.query(ServerGroup):
        for address in (sg.master_address, sg.master_alt_address):
            if not address:
                continue
            if address in configured_addresses:
                this_servers_addresses.append(address)
    
    # Calculate master_dns_server
    master_sm = get_master_sm(db_session)
    replica_sg = master_sm.replica_sg
    master_address = None
    master_server = None
    if replica_sg:
        if replica_sg.master_address:
            if replica_sg.master_address in this_servers_addresses:
                master_address = replica_sg.master_address
        if replica_sg.master_alt_address:
            if replica_sg.master_alt_address in this_servers_addresses:
                master_address = replica_sg.master_alt_address
    # Prefer master_address for settings['master_dns_server']
    if master_address:
        settings['master_dns_server'] = master_address
        # Recalculate master server
        for server_sm in replica_sg.servers:
            if server_sm.address == master_address:
                master_server = server_sm
    master_sm.master_server = master_server
    master_sm.master_hostname = socket.gethostname()

    # sort | uniq all the addresses to remove any literal duplicates
    this_servers_addresses.append(settings['master_dns_server'])
    this_servers_addresses = list(set(sorted(this_servers_addresses)))
    settings['this_servers_addresses'] = this_servers_addresses


